#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
"""
Usage:

    # module.py
    import mockable

    @mockable.mockable
    def expensive_function():
        pass

    def function_a():
        mockable.expensive_function()


    # test_module.py
    import pytest

    @pytest.fixture
    def mock_expensive_function(monkeypatch):
        def cheap_func():
            pass
        monkeypatch.setattr(
            'mockable.expensive_function',
            cheap_func
        )

    def test_function_a(mock_expensive_function):
        function_a()

"""

import sys


def mockable():
    def decorator(func):
        setattr(sys.modules[__name__], func.__name__, func)
        return func

    return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import json

from .connections import metadata_db_connection_proxy
from .exceptions import UnrecognizedUSAFIDError, UnrecognizedZCTAError

from .utils import lazy_property
from .validation import valid_zcta_or_raise


__all__ = ("get_lat_long_climate_zones", "get_zcta_metadata", "zcta_to_lat_long")


class CachedData(object):
    @lazy_property
    def climate_zone_geometry(self):
        try:
            from shapely.geometry import shape
        except ImportError:  # pragma: no cover
            raise ImportError(
                "Matching by lat/lng within climate zone requires shapely"
            )

        conn = metadata_db_connection_proxy.get_connection()
        cur = conn.cursor()

        cur.execute(
            """
          select
            iecc_climate_zone, geometry
          from
            iecc_climate_zone_metadata
        """
        )
        iecc_climate_zones = [
            (cz_id, shape(json.loads(geometry))) for (cz_id, geometry) in cur.fetchall()
        ]

        cur.execute(
            """
          select
            iecc_moisture_regime, geometry
          from
            iecc_moisture_regime_metadata
        """
        )
        iecc_moisture_regimes = [
            (cz_id, shape(json.loads(geometry))) for (cz_id, geometry) in cur.fetchall()
        ]

        cur.execute(
            """
          select
            ba_climate_zone, geometry
          from
            ba_climate_zone_metadata
        """
        )
        ba_climate_zones = [
            (cz_id, shape(json.loads(geometry))) for (cz_id, geometry) in cur.fetchall()
        ]

        cur.execute(
            """
          select
            ca_climate_zone, geometry
          from
            ca_climate_zone_metadata
        """
        )
        ca_climate_zones = [
            (cz_id, shape(json.loads(geometry))) for (cz_id, geometry) in cur.fetchall()
        ]

        return (
            iecc_climate_zones,
            iecc_moisture_regimes,
            ba_climate_zones,
            ca_climate_zones,
        )


cached_data = CachedData()


def get_lat_long_climate_zones(latitude, longitude):
    """ Get climate zones that contain lat/long coordinates.

    Parameters
    ----------
    latitude : float
        Latitude of point.
    longitude : float
        Longitude of point.

    Returns
    -------
    climate_zones: dict of str
        Region ids for each climate zone type.
    """
    try:
        from shapely.geometry import Point
    except ImportError:  # pragma: no cover
        raise ImportError("Finding climate zone of lat/long points requires shapely.")

    (
        iecc_climate_zones,
        iecc_moisture_regimes,
        ba_climate_zones,
        ca_climate_zones,
    ) = cached_data.climate_zone_geometry

    point = Point(longitude, latitude)  # x,y
    climate_zones = {}
    for iecc_climate_zone, shape in iecc_climate_zones:
        if shape.contains(point):
            climate_zones["iecc_climate_zone"] = iecc_climate_zone
            break
    else:
        climate_zones["iecc_climate_zone"] = None

    for iecc_moisture_regime, shape in iecc_moisture_regimes:
        if shape.contains(point):
            climate_zones["iecc_moisture_regime"] = iecc_moisture_regime
            break
    else:
        climate_zones["iecc_moisture_regime"] = None

    for ba_climate_zone, shape in ba_climate_zones:
        if shape.contains(point):
            climate_zones["ba_climate_zone"] = ba_climate_zone
            break
    else:
        climate_zones["ba_climate_zone"] = None

    for ca_climate_zone, shape in ca_climate_zones:
        if shape.contains(point):
            climate_zones["ca_climate_zone"] = ca_climate_zone
            break
    else:
        climate_zones["ca_climate_zone"] = None

    return climate_zones


def get_zcta_metadata(zcta):
    """ Get metadata about a ZIP Code Tabulation Area (ZCTA).

    Parameters
    ----------
    zcta : str
        ID of ZIP Code Tabulation Area

    Returns
    -------
    metadata : dict
        Dict of data about the ZCTA, including lat/long coordinates.
    """
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()
    cur.execute(
        """
      select
        *
      from
        zcta_metadata
      where
        zcta_id = ?
    """,
        (zcta,),
    )
    row = cur.fetchone()
    if row is None:
        raise UnrecognizedZCTAError(zcta)
    return {col[0]: row[i] for i, col in enumerate(cur.description)}


def zcta_to_lat_long(zcta):
    """Get location of ZCTA centroid

    Retrieves latitude and longitude of centroid of ZCTA
    to use for matching with weather station.

    Parameters
    ----------
    zcta : str
        ID of the target ZCTA.

    Returns
    -------
    latitude : float
        Latitude of centroid of ZCTA.
    longitude : float
        Target Longitude of centroid of ZCTA.
    """
    valid_zcta_or_raise(zcta)

    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()

    cur.execute(
        """
      select
        latitude
        , longitude
      from
        zcta_metadata
      where
        zcta_id = ?
    """,
        (zcta,),
    )
    # match existence checked in validate_zcta_or_raise(zcta)
    latitude, longitude = cur.fetchone()

    return float(latitude), float(longitude)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import os
import json

try:
    from sqlalchemy import create_engine, MetaData, Table, Column, String, DateTime
    from sqlalchemy.sql import select, func
    from sqlalchemy.exc import IntegrityError
except ImportError:  # pragma: no cover
    has_sqlalchemy = False
else:
    has_sqlalchemy = True
import pytz


def get_datetime_if_exists(data):
    if data is None:
        return None
    else:
        dt = data[0]
    if is_tz_naive(dt):
        return pytz.UTC.localize(data[0])
    else:
        return dt


def is_tz_naive(dt):
    return dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None


class KeyValueStore(object):
    def __init__(self, url=None):
        if not has_sqlalchemy:  # pragma: no cover
            raise ImportError("KeyValueStore requires sqlalchemy.")
        self._prepare_db(url)

    def __repr__(self):
        return 'KeyValueStore("{}")'.format(self.url)

    def _get_url(self):  # pragma: no cover (tests always provide url)
        url = os.environ.get("EEWEATHER_CACHE_URL")
        if url is None:
            directory = "{}/.eeweather".format(os.path.expanduser("~"))
            if not os.path.exists(directory):
                os.makedirs(directory)
            url = "sqlite:///{}/cache.db".format(directory)
        return url

    def _prepare_db(self, url=None):

        # set url
        if url is None:  # pragma: no cover (tests always provide url)
            url = self._get_url()
        self.url = url

        eng = create_engine(url)
        metadata = MetaData(eng)

        tbl_items = Table(
            "items",
            metadata,
            Column("key", String, unique=True, index=True),  # arbitrary unique key
            Column("data", String),  # arbitrary json
            Column("updated", DateTime(timezone=True)),  # time of last transaction
        )

        # only create if not already created
        tbl_items.create(checkfirst=True)

        self.items = tbl_items

    def key_exists(self, key):
        s = select([self.items.c.key]).where(self.items.c.key == key)
        result = s.execute()
        return result.fetchone() is not None

    def save_json(self, key, data):
        data = json.dumps(data, separators=(",", ":"))
        updated = func.now()
        try:
            s = self.items.insert().values(key=key, data=data, updated=updated)
            s.execute()
        except IntegrityError:
            s = (
                self.items.update()
                .where(self.items.c.key == key)
                .values(key=key, data=data, updated=updated)
            )
            s.execute()

    def retrieve_json(self, key):
        s = select([self.items.c.data]).where(self.items.c.key == key)
        result = s.execute()
        data = result.fetchone()
        if data is None:
            return None
        else:
            return json.loads(data[0])

    def key_updated(self, key):
        s = select([self.items.c.updated]).where(self.items.c.key == key)
        result = s.execute()
        data = result.fetchone()
        return get_datetime_if_exists(data)

    def clear(self, key=None):
        if key is None:
            s = self.items.delete()
        else:
            s = self.items.delete().where(self.items.c.key == key)
        s.execute()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from .connections import metadata_db_connection_proxy
from .exceptions import UnrecognizedZCTAError, UnrecognizedUSAFIDError


__all__ = ("valid_zcta_or_raise", "valid_usaf_id_or_raise")


def valid_zcta_or_raise(zcta):
    """ Check if ZCTA is valid and raise eeweather.UnrecognizedZCTAError if not. """
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()

    cur.execute(
        """
      select exists (
        select
          zcta_id
        from
          zcta_metadata
        where
          zcta_id = ?
      )
    """,
        (zcta,),
    )
    (exists,) = cur.fetchone()
    if exists:
        return True
    else:
        raise UnrecognizedZCTAError(zcta)


def valid_usaf_id_or_raise(usaf_id):
    """ Check if USAF ID is valid and raise eeweather.UnrecognizedUSAFIDError if not. """
    conn = metadata_db_connection_proxy.get_connection()
    cur = conn.cursor()

    cur.execute(
        """
      select exists (
        select
          usaf_id
        from
          isd_station_metadata
        where
          usaf_id = ?
      )
    """,
        (usaf_id,),
    )
    (exists,) = cur.fetchone()
    if exists:
        return True
    else:
        raise UnrecognizedUSAFIDError(usaf_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import pytest

from eeweather import (
    EEWeatherError,
    UnrecognizedUSAFIDError,
    UnrecognizedZCTAError,
    ISDDataNotAvailableError,
    GSODDataNotAvailableError,
)


def test_eeweather_error():
    with pytest.raises(EEWeatherError) as excinfo:
        raise EEWeatherError("message")
    assert excinfo.value.args[0] == "message"


def test_unrecognized_usaf_id_error():
    with pytest.raises(UnrecognizedUSAFIDError) as excinfo:
        raise UnrecognizedUSAFIDError("INVALID")
    assert excinfo.value.value == "INVALID"
    assert excinfo.value.message == (
        'The value "INVALID" was not recognized as a'
        " valid USAF weather station identifier."
    )


def test_unrecognized_zcta_error():
    with pytest.raises(UnrecognizedZCTAError) as excinfo:
        raise UnrecognizedZCTAError("INVALID")
    assert excinfo.value.value == "INVALID"
    assert excinfo.value.message == (
        'The value "INVALID" was not recognized as a valid ZCTA identifier.'
    )


def test_isd_data_does_not_exist_error():
    with pytest.raises(ISDDataNotAvailableError) as excinfo:
        raise ISDDataNotAvailableError("123456", 1800)
    assert excinfo.value.usaf_id == "123456"
    assert excinfo.value.year == 1800
    assert excinfo.value.message == (
        'ISD data does not exist for station "123456" in year 1800.'
    )


def test_gsod_data_does_not_exist_error():
    with pytest.raises(GSODDataNotAvailableError) as excinfo:
        raise GSODDataNotAvailableError("123456", 1800)
    assert excinfo.value.usaf_id == "123456"
    assert excinfo.value.year == 1800
    assert excinfo.value.message == (
        'GSOD data does not exist for station "123456" in year 1800.'
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from datetime import datetime
import pandas as pd
import pytest
import pytz

from eeweather import rank_stations, combine_ranked_stations, select_station
from eeweather.exceptions import ISDDataNotAvailableError


@pytest.fixture
def lat_long_fresno():
    return 36.7378, -119.7871


@pytest.fixture
def lat_long_africa():
    return 0, 0


def test_rank_stations_no_filter(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno
    df = rank_stations(lat, lng)
    snapshot.assert_match(df.shape, "df.shape")
    assert list(df.columns) == [
        "rank",
        "distance_meters",
        "latitude",
        "longitude",
        "iecc_climate_zone",
        "iecc_moisture_regime",
        "ba_climate_zone",
        "ca_climate_zone",
        "rough_quality",
        "elevation",
        "state",
        "tmy3_class",
        "is_tmy3",
        "is_cz2010",
        "difference_elevation_meters",
    ]
    assert round(df.distance_meters.iloc[0]) == 3008.0
    assert round(df.distance_meters.iloc[-10]) == 9614514.0
    assert pd.isnull(df.distance_meters.iloc[-1]) is True


def test_rank_stations_match_climate_zones_not_null(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno
    df = rank_stations(lat, lng, match_iecc_climate_zone=True)
    snapshot.assert_match(df.shape, "match_iecc_climate_zone")

    df = rank_stations(lat, lng, match_iecc_moisture_regime=True)
    snapshot.assert_match(df.shape, "match_iecc_moisture_regime")

    df = rank_stations(lat, lng, match_ba_climate_zone=True)
    snapshot.assert_match(df.shape, "match_ba_climate_zone")

    df = rank_stations(lat, lng, match_ca_climate_zone=True)
    snapshot.assert_match(df.shape, "match_ca_climate_zone")


def test_rank_stations_match_climate_zones_null(lat_long_africa, snapshot):
    lat, lng = lat_long_africa
    df = rank_stations(lat, lng, match_iecc_climate_zone=True)
    snapshot.assert_match(df.shape, "match_iecc_climate_zone")

    df = rank_stations(lat, lng, match_iecc_moisture_regime=True)
    snapshot.assert_match(df.shape, "match_iecc_moisture_regime")

    df = rank_stations(lat, lng, match_ba_climate_zone=True)
    snapshot.assert_match(df.shape, "match_ba_climate_zone")

    df = rank_stations(lat, lng, match_ca_climate_zone=True)
    snapshot.assert_match(df.shape, "match_ca_climate_zone")


def test_rank_stations_match_state(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno
    df = rank_stations(lat, lng, site_state="CA")
    snapshot.assert_match(df.shape, "site_state=CA, match_state=False")

    df = rank_stations(lat, lng, site_state="CA", match_state=True)
    snapshot.assert_match(df.shape, "site_state=CA, match_state=True")

    df = rank_stations(lat, lng, site_state=None, match_state=True)
    snapshot.assert_match(df.shape, "site_state=None, match_state=True")


def test_rank_stations_is_tmy3(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno
    df = rank_stations(lat, lng, is_tmy3=True)
    snapshot.assert_match(df.shape, "is_tmy3=True")

    df = rank_stations(lat, lng, is_tmy3=False)
    snapshot.assert_match(df.shape, "is_tmy3=False")


def test_rank_stations_is_cz2010(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno
    df = rank_stations(lat, lng, is_cz2010=True)
    snapshot.assert_match(df.shape, "is_cz2010=True")

    df = rank_stations(lat, lng, is_cz2010=False)
    snapshot.assert_match(df.shape, "is_cz2010=False")


def test_rank_stations_minimum_quality(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno
    df = rank_stations(lat, lng, minimum_quality="low")
    snapshot.assert_match(df.shape, "minimum_quality=low")

    df = rank_stations(lat, lng, minimum_quality="medium")
    snapshot.assert_match(df.shape, "minimum_quality=medium")

    df = rank_stations(lat, lng, minimum_quality="high")
    snapshot.assert_match(df.shape, "minimum_quality=high")


def test_rank_stations_minimum_tmy3_class(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno
    df = rank_stations(lat, lng, minimum_tmy3_class="III")
    snapshot.assert_match(df.shape, "minimum_tmy3_class=III")

    df = rank_stations(lat, lng, minimum_tmy3_class="II")
    snapshot.assert_match(df.shape, "minimum_tmy3_class=II")

    df = rank_stations(lat, lng, minimum_tmy3_class="I")
    snapshot.assert_match(df.shape, "minimum_tmy3_class=I")


def test_rank_stations_max_distance_meters(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno

    df = rank_stations(lat, lng, max_distance_meters=200000)
    snapshot.assert_match(df.shape, "max_distance_meters=200000")

    df = rank_stations(lat, lng, max_distance_meters=50000)
    snapshot.assert_match(df.shape, "max_distance_meters=50000")


def test_rank_stations_max_difference_elevation_meters(lat_long_fresno, snapshot):
    lat, lng = lat_long_fresno

    # no site_elevation
    df = rank_stations(lat, lng, max_difference_elevation_meters=200)
    snapshot.assert_match(df.shape, "max_difference_elevation_meters=200")

    df = rank_stations(lat, lng, site_elevation=0, max_difference_elevation_meters=200)
    snapshot.assert_match(
        df.shape, "site_elevation=0, max_difference_elevation_meters=200"
    )

    df = rank_stations(lat, lng, site_elevation=0, max_difference_elevation_meters=50)
    snapshot.assert_match(
        df.shape, "site_elevation=0, max_difference_elevation_meters=50"
    )

    df = rank_stations(
        lat, lng, site_elevation=1000, max_difference_elevation_meters=50
    )
    snapshot.assert_match(
        df.shape, "site_elevation=1000, max_difference_elevation_meters=50"
    )


@pytest.fixture
def cz_candidates(lat_long_fresno):
    lat, lng = lat_long_fresno
    return rank_stations(
        lat,
        lng,
        match_iecc_climate_zone=True,
        match_iecc_moisture_regime=True,
        match_ba_climate_zone=True,
        match_ca_climate_zone=True,
        minimum_quality="high",
        is_tmy3=True,
        is_cz2010=True,
    )


@pytest.fixture
def naive_candidates(lat_long_fresno):
    lat, lng = lat_long_fresno
    return rank_stations(
        lat, lng, minimum_quality="high", is_tmy3=True, is_cz2010=True
    ).head()


def test_combine_ranked_stations_empty():
    with pytest.raises(ValueError):
        combine_ranked_stations([])


def test_combine_ranked_stations(cz_candidates, naive_candidates):
    assert list(cz_candidates.index) == ["723890", "723896", "723895", "723840"]
    assert list(naive_candidates.index) == [
        "723890",
        "723896",
        "724815",
        "723895",
        "723965",
    ]

    combined_candidates = combine_ranked_stations([cz_candidates, naive_candidates])

    assert combined_candidates.shape == (6, 15)
    assert combined_candidates["rank"].iloc[0] == 1
    assert combined_candidates["rank"].iloc[-1] == 6
    assert list(combined_candidates.index) == [
        "723890",
        "723896",
        "723895",
        "723840",
        "724815",
        "723965",
    ]


def test_select_station_no_coverage_check(cz_candidates):
    station, warnings = select_station(cz_candidates)
    assert station.usaf_id == "723890"


@pytest.fixture
def monkeypatch_load_isd_hourly_temp_data(monkeypatch):
    def load_isd_hourly_temp_data(station, start, end, fetch_from_web=True):
        # because result datetimes should fall exactly on hours
        normalized_start = datetime(
            start.year, start.month, start.day, start.hour, tzinfo=pytz.UTC
        )
        normalized_end = datetime(
            end.year, end.month, end.day, end.hour, tzinfo=pytz.UTC
        )
        index = pd.date_range(normalized_start, normalized_end, freq="H", tz="UTC")

        # simulate missing data
        if station.usaf_id in ("723890", "723896"):
            return pd.Series(1, index=index)[: -24 * 50].reindex(index), []
        return pd.Series(1, index=index)[: -24 * 10].reindex(index), []

    monkeypatch.setattr(
        "eeweather.mockable.load_isd_hourly_temp_data", load_isd_hourly_temp_data
    )


def test_select_station_full_data(cz_candidates, monkeypatch_load_isd_hourly_temp_data):
    start = datetime(2017, 1, 1, tzinfo=pytz.UTC)
    end = datetime(2018, 1, 1, tzinfo=pytz.UTC)

    # 1st misses qualification
    station, warnings = select_station(cz_candidates, coverage_range=(start, end))
    assert station.usaf_id == "723895"

    # 1st meets qualification
    station, warnings = select_station(
        cz_candidates, coverage_range=(start, end), min_fraction_coverage=0.8
    )
    assert station.usaf_id == "723890"

    # none meet qualification
    station, warnings = select_station(
        cz_candidates, coverage_range=(start, end), min_fraction_coverage=0.99
    )
    assert station is None


@pytest.fixture
def monkeypatch_load_isd_hourly_temp_data_with_error(monkeypatch):
    def load_isd_hourly_temp_data(station, start, end, fetch_from_web=True):
        index = pd.date_range(start, end, freq="H", tz="UTC")
        if station.usaf_id == "723890":
            raise ISDDataNotAvailableError(
                "723890", start.year
            )  # first choice not available
        elif station.usaf_id == "723896":
            return pd.Series(1, index=index)[: -24 * 10].reindex(index), []
        else:  # pragma: no cover - only for helping to debug failing tests
            raise ValueError(
                "The requested station is not specified in the monkeypatched data: {}.".format(
                    station
                )
            )

    monkeypatch.setattr(
        "eeweather.mockable.load_isd_hourly_temp_data", load_isd_hourly_temp_data
    )


def test_select_station_with_isd_data_not_available_error(
    cz_candidates, monkeypatch_load_isd_hourly_temp_data_with_error
):
    start = datetime(2017, 1, 1, tzinfo=pytz.UTC)
    end = datetime(2018, 1, 1, tzinfo=pytz.UTC)

    # 1st misses qualification because data not available
    station, warnings = select_station(
        cz_candidates, coverage_range=(start, end), min_fraction_coverage=0.8
    )
    assert station.usaf_id == "723896"


@pytest.fixture
def monkeypatch_load_isd_hourly_temp_data_with_empty(monkeypatch):
    def load_isd_hourly_temp_data(station, start, end, fetch_from_web=True):
        index = pd.date_range(start, end, freq="H", tz="UTC")
        if station.usaf_id == "723890":
            return pd.Series(1, index=index)[:0], []
        elif station.usaf_id == "723896":
            return pd.Series(1, index=index)[: -24 * 10].reindex(index), []
        else:  # pragma: no cover - only for helping to debug failing tests
            raise ValueError(
                "The requested station is not specified in the monkeypatched data: {}.".format(
                    station
                )
            )

    monkeypatch.setattr(
        "eeweather.mockable.load_isd_hourly_temp_data", load_isd_hourly_temp_data
    )


def test_select_station_with_empty_tempC(
    cz_candidates, monkeypatch_load_isd_hourly_temp_data_with_empty, snapshot
):
    start = datetime(2017, 1, 1, tzinfo=pytz.UTC)
    end = datetime(2018, 1, 1, tzinfo=pytz.UTC)

    # 1st misses qualification because data not available
    station, warnings = select_station(
        cz_candidates, coverage_range=(start, end), min_fraction_coverage=0.8
    )
    snapshot.assert_match(station.usaf_id, "station_id")


def test_select_station_distance_warnings_check(lat_long_africa):
    lat, lng = lat_long_africa
    df = rank_stations(lat, lng)
    station, warnings = select_station(df)
    assert len(warnings) == 2
    assert warnings[0].qualified_name == "eeweather.exceeds_maximum_distance"
    assert warnings[1].qualified_name == "eeweather.exceeds_maximum_distance"
    assert warnings[0].data["max_distance_meters"] == 50000
    assert warnings[1].data["max_distance_meters"] == 200000


def test_select_station_no_station_warnings_check():
    df = pd.DataFrame()
    station, warnings = select_station(df)
    assert warnings[0].qualified_name == "eeweather.no_weather_station_selected"
    assert warnings[0].data == {"rank": 1, "min_fraction_coverage": 0.9}


def test_select_station_with_second_level_dates(
    cz_candidates, monkeypatch_load_isd_hourly_temp_data, snapshot
):
    # dates don't fall exactly on the hour
    start = datetime(2017, 1, 1, 2, 3, 4, tzinfo=pytz.UTC)
    end = datetime(2018, 1, 1, 12, 13, 14, tzinfo=pytz.UTC)

    station, warnings = select_station(cz_candidates, coverage_range=(start, end))
    snapshot.assert_match(station.usaf_id, "station_id")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import pytest

from eeweather import get_version
from eeweather.geo import (
    get_lat_long_climate_zones,
    get_zcta_metadata,
    zcta_to_lat_long,
)
from eeweather.exceptions import UnrecognizedZCTAError, UnrecognizedUSAFIDError


def test_get_version():
    assert get_version() is not None


def test_get_zcta_metadata():
    metadata = get_zcta_metadata("90006")
    assert metadata == {
        "ba_climate_zone": "Hot-Dry",
        "ca_climate_zone": "CA_09",
        "geometry": None,
        "iecc_climate_zone": "3",
        "iecc_moisture_regime": "B",
        "latitude": "34.0480061236777",
        "longitude": "-118.294181179333",
        "state": "CA",
        "zcta_id": "90006",
    }

    with pytest.raises(UnrecognizedZCTAError) as excinfo:
        get_zcta_metadata("00000")
    assert excinfo.value.value == "00000"


def test_get_lat_long_climate_zones():
    climate_zones = get_lat_long_climate_zones(35.1, -119.2)
    assert climate_zones == {
        "iecc_climate_zone": "3",
        "iecc_moisture_regime": "B",
        "ba_climate_zone": "Hot-Dry",
        "ca_climate_zone": "CA_13",
    }


def test_get_lat_long_climate_zones_out_of_range():
    climate_zones = get_lat_long_climate_zones(0, 0)
    assert climate_zones == {
        "iecc_climate_zone": None,
        "iecc_moisture_regime": None,
        "ba_climate_zone": None,
        "ca_climate_zone": None,
    }


def test_zcta_to_lat_long():
    with pytest.raises(UnrecognizedZCTAError) as excinfo:
        zcta_to_lat_long("00000")

    lat, lng = zcta_to_lat_long("70001")
    assert round(lat) == 30
    assert round(lng) == -90

    lat, lng = zcta_to_lat_long("94574")
    assert round(lat) == 39
    assert round(lng) == -122



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_get_zcta_ids n_zcta_ids"] = 33144

snapshots["test_get_zcta_ids_by_state n_zcta_ids"] = 1763

snapshots["test_get_isd_station_usaf_ids n_usaf_ids"] = 3874

snapshots["test_get_isd_station_usaf_ids_by_state n_usaf_ids"] = 77



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from eeweather.utils import lazy_property


def test_lazy_property():
    class NotLazy(object):

        n_times = 1

        @property
        def tryme(self):
            times = self.n_times
            self.n_times += 1
            return times

    class Lazy(object):

        n_times = 1

        @lazy_property
        def tryme(self):
            times = self.n_times
            self.n_times += 1
            return times

    not_lazy = NotLazy()
    assert not_lazy.tryme == 1
    assert not_lazy.tryme == 2  # called twice

    lazy = Lazy()
    assert lazy.tryme == 1
    assert lazy.tryme == 1  # only called once



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import pandas as pd
import numpy as np
import pyproj

import eeweather.mockable
from .exceptions import ISDDataNotAvailableError
from .connections import metadata_db_connection_proxy
from .geo import get_lat_long_climate_zones
from .stations import ISDStation
from .utils import lazy_property
from .warnings import EEWeatherWarning

__all__ = ("rank_stations", "combine_ranked_stations", "select_station")


class CachedData(object):
    @lazy_property
    def all_station_metadata(self):
        conn = metadata_db_connection_proxy.get_connection()
        cur = conn.cursor()
        cur.execute(
            """
          select
            isd.usaf_id
            , isd.latitude
            , isd.longitude
            , isd.iecc_climate_zone
            , isd.iecc_moisture_regime
            , isd.ba_climate_zone
            , isd.ca_climate_zone
            , isd.quality as rough_quality
            , isd.elevation
            , isd.state
            , tmy3.class as tmy3_class
            , tmy3.usaf_id is not null as is_tmy3
            , cz2010.usaf_id is not null as is_cz2010
          from
            isd_station_metadata as isd
            left join cz2010_station_metadata as cz2010 on
              isd.usaf_id = cz2010.usaf_id
            left join tmy3_station_metadata as tmy3 on
              isd.usaf_id = tmy3.usaf_id
          order by
            isd.usaf_id
        """
        )

        df = pd.DataFrame(
            [
                {col[0]: val for col, val in zip(cur.description, row)}
                for row in cur.fetchall()
            ],
            columns=[
                "usaf_id",
                "latitude",
                "longitude",
                "iecc_climate_zone",
                "iecc_moisture_regime",
                "ba_climate_zone",
                "ca_climate_zone",
                "rough_quality",
                "elevation",
                "state",
                "tmy3_class",
                "is_tmy3",
                "is_cz2010",
            ],
        ).set_index("usaf_id")

        df["latitude"] = df.latitude.astype(float)
        df["longitude"] = df.longitude.astype(float)
        df["elevation"] = df.elevation.astype(float)
        df["is_tmy3"] = df.is_tmy3.astype(bool)
        df["is_cz2010"] = df.is_cz2010.astype(bool)
        return df


cached_data = CachedData()


def _combine_filters(filters, index):
    combined_filters = pd.Series(True, index=index)
    for f in filters:
        combined_filters &= f
    return combined_filters


def rank_stations(
    site_latitude,
    site_longitude,
    site_state=None,
    site_elevation=None,
    match_iecc_climate_zone=False,
    match_iecc_moisture_regime=False,
    match_ba_climate_zone=False,
    match_ca_climate_zone=False,
    match_state=False,
    minimum_quality=None,
    minimum_tmy3_class=None,
    max_distance_meters=None,
    max_difference_elevation_meters=None,
    is_tmy3=None,
    is_cz2010=None,
):
    """ Get a ranked, filtered set of candidate weather stations and metadata
    for a particular site.

    Parameters
    ----------
    site_latitude : float
        Latitude of target site for which to find candidate weather stations.
    site_longitude : float
        Longitude of target site for which to find candidate weather stations.
    site_state : str, 2 letter abbreviation
        US state of target site, used optionally to filter potential candidate
        weather stations. Ignored unless ``match_state=True``.
    site_elevation : float
        Elevation of target site in meters, used optionally to filter potential
        candidate weather stations. Ignored unless
        ``max_difference_elevation_meters`` is set.
    match_iecc_climate_zone : bool
        If ``True``, filter candidate weather stations to those
        matching the IECC climate zone of the target site.
    match_iecc_moisture_regime : bool
        If ``True``, filter candidate weather stations to those
        matching the IECC moisture regime of the target site.
    match_ca_climate_zone : bool
        If ``True``, filter candidate weather stations to those
        matching the CA climate zone of the target site.
    match_ba_climate_zone : bool
        If ``True``, filter candidate weather stations to those
        matching the Building America climate zone of the target site.
    match_state : bool
        If ``True``, filter candidate weather stations to those
        matching the US state of the target site, as specified by
        ``site_state=True``.
    minimum_quality : str, ``'high'``, ``'medium'``, ``'low'``
        If given, filter candidate weather stations to those meeting or
        exceeding the given quality, as summarized by the frequency and
        availability of observations in the NOAA Integrated Surface Database.
    minimum_tmy3_class : str, ``'I'``, ``'II'``, ``'III'``
        If given, filter candidate weather stations to those meeting or
        exceeding the given class, as reported in the NREL TMY3 metadata.
    max_distance_meters : float
        If given, filter candidate weather stations to those within the
        ``max_distance_meters`` of the target site location.
    max_difference_elevation_meters : float
        If given, filter candidate weather stations to those with elevations
        within ``max_difference_elevation_meters`` of the target site elevation.
    is_tmy3 : bool
        If given, filter candidate weather stations to those for which TMY3
        normal year temperature data is available.
    is_cz2010 : bool
        If given, filter candidate weather stations to those for which CZ2010
        normal year temperature data is available.

    Returns
    -------
    ranked_filtered_candidates : :any:`pandas.DataFrame`
        Index is ``usaf_id``.  Each row contains a potential weather station
        match and metadata. Contains the following columns:

        - ``rank``: Rank of weather station match for the target site.
        - ``distance_meters``: Distance from target site to weather station site.
        - ``latitude``: Latitude of weather station site.
        - ``longitude``: Longitude of weather station site.
        - ``iecc_climate_zone``: IECC Climate Zone ID (1-8)
        - ``iecc_moisture_regime``: IECC Moisture Regime ID (A-C)
        - ``ba_climate_zone``: Building America climate zone name
        - ``ca_climate_zone``: Califoria climate zone number
        - ``rough_quality``: Approximate measure of frequency of ISD
          observations data at weather station.
        - ``elevation``: Elevation of weather station site, if available.
        - ``state``: US state of weather station site, if applicable.
        - ``tmy3_class``: Weather station class as reported by NREL TMY3, if
          available
        - ``is_tmy3``: Weather station has associated TMY3 data.
        - ``is_cz2010``: Weather station has associated CZ2010 data.
        - ``difference_elevation_meters``: Absolute difference in meters
          between target site elevation and weather station elevation, if
          available.

    """
    candidates = cached_data.all_station_metadata

    # compute distances
    candidates_defined_lat_long = candidates[
        candidates.latitude.notnull() & candidates.longitude.notnull()
    ]
    candidates_latitude = candidates_defined_lat_long.latitude
    candidates_longitude = candidates_defined_lat_long.longitude
    tiled_site_latitude = np.tile(site_latitude, candidates_latitude.shape)
    tiled_site_longitude = np.tile(site_longitude, candidates_longitude.shape)
    geod = pyproj.Geod(ellps="WGS84")
    dists = geod.inv(
        tiled_site_longitude,
        tiled_site_latitude,
        candidates_longitude.values,
        candidates_latitude.values,
    )[2]
    distance_meters = pd.Series(dists, index=candidates_defined_lat_long.index).reindex(
        candidates.index
    )
    candidates["distance_meters"] = distance_meters

    if site_elevation is not None:
        difference_elevation_meters = (candidates.elevation - site_elevation).abs()
    else:
        difference_elevation_meters = None
    candidates["difference_elevation_meters"] = difference_elevation_meters

    site_climate_zones = get_lat_long_climate_zones(site_latitude, site_longitude)
    site_iecc_climate_zone = site_climate_zones["iecc_climate_zone"]
    site_iecc_moisture_regime = site_climate_zones["iecc_moisture_regime"]
    site_ca_climate_zone = site_climate_zones["ca_climate_zone"]
    site_ba_climate_zone = site_climate_zones["ba_climate_zone"]

    # create filters
    filters = []

    if match_iecc_climate_zone:
        if site_iecc_climate_zone is None:
            filters.append(candidates.iecc_climate_zone.isnull())
        else:
            filters.append(candidates.iecc_climate_zone == site_iecc_climate_zone)
    if match_iecc_moisture_regime:
        if site_iecc_moisture_regime is None:
            filters.append(candidates.iecc_moisture_regime.isnull())
        else:
            filters.append(candidates.iecc_moisture_regime == site_iecc_moisture_regime)
    if match_ba_climate_zone:
        if site_ba_climate_zone is None:
            filters.append(candidates.ba_climate_zone.isnull())
        else:
            filters.append(candidates.ba_climate_zone == site_ba_climate_zone)
    if match_ca_climate_zone:
        if site_ca_climate_zone is None:
            filters.append(candidates.ca_climate_zone.isnull())
        else:
            filters.append(candidates.ca_climate_zone == site_ca_climate_zone)

    if match_state:
        if site_state is None:
            filters.append(candidates.state.isnull())
        else:
            filters.append(candidates.state == site_state)

    if is_tmy3 is not None:
        filters.append(candidates.is_tmy3.isin([is_tmy3]))
    if is_cz2010 is not None:
        filters.append(candidates.is_cz2010.isin([is_cz2010]))

    if minimum_quality == "low":
        filters.append(candidates.rough_quality.isin(["high", "medium", "low"]))
    elif minimum_quality == "medium":
        filters.append(candidates.rough_quality.isin(["high", "medium"]))
    elif minimum_quality == "high":
        filters.append(candidates.rough_quality.isin(["high"]))

    if minimum_tmy3_class == "III":
        filters.append(candidates.tmy3_class.isin(["I", "II", "III"]))
    elif minimum_tmy3_class == "II":
        filters.append(candidates.tmy3_class.isin(["I", "II"]))
    elif minimum_tmy3_class == "I":
        filters.append(candidates.tmy3_class.isin(["I"]))

    if max_distance_meters is not None:
        filters.append(candidates.distance_meters <= max_distance_meters)

    if max_difference_elevation_meters is not None and site_elevation is not None:
        filters.append(
            candidates.difference_elevation_meters <= max_difference_elevation_meters
        )

    combined_filters = _combine_filters(filters, candidates.index)
    filtered_candidates = candidates[combined_filters]
    ranked_filtered_candidates = filtered_candidates.sort_values(by=["distance_meters"])

    # add rank column
    ranks = range(1, 1 + len(ranked_filtered_candidates))
    ranked_filtered_candidates.insert(0, "rank", ranks)

    return ranked_filtered_candidates[
        [
            "rank",
            "distance_meters",
            "latitude",
            "longitude",
            "iecc_climate_zone",
            "iecc_moisture_regime",
            "ba_climate_zone",
            "ca_climate_zone",
            "rough_quality",
            "elevation",
            "state",
            "tmy3_class",
            "is_tmy3",
            "is_cz2010",
            "difference_elevation_meters",
        ]
    ]


def combine_ranked_stations(rankings):
    """ Combine :any:`pandas.DataFrame` s of candidate weather stations to form
    a hybrid ranking dataframe.

    Parameters
    ----------
    rankings : list of :any:`pandas.DataFrame`
        Dataframes of ranked weather station candidates and metadata.
        All ranking dataframes should have the same columns and must be
        sorted by rank.

    Returns
    -------
    ranked_filtered_candidates : :any:`pandas.DataFrame`
        Dataframe has a rank column and the same columns given in the source
        dataframes.
    """

    if len(rankings) == 0:
        raise ValueError("Requires at least one ranking.")

    combined_ranking = rankings[0]
    for ranking in rankings[1:]:
        filtered_ranking = ranking[~ranking.index.isin(combined_ranking.index)]
        combined_ranking = pd.concat([combined_ranking, filtered_ranking])

    combined_ranking["rank"] = range(1, 1 + len(combined_ranking))
    return combined_ranking


@eeweather.mockable.mockable()
def load_isd_hourly_temp_data(
    station, start_date, end_date, fetch_from_web
):  # pragma: no cover
    return station.load_isd_hourly_temp_data(
        start_date, end_date, fetch_from_web=fetch_from_web
    )


def select_station(
    candidates,
    coverage_range=None,
    min_fraction_coverage=0.9,
    distance_warnings=(50000, 200000),
    rank=1,
    fetch_from_web=True,
):
    """ Select a station from a list of candidates that meets given data
    quality criteria.

    Parameters
    ----------
    candidates : :any:`pandas.DataFrame`
        A dataframe of the form given by :any:`eeweather.rank_stations` or
        :any:`eeweather.combine_ranked_stations`, specifically having at least
        an index with ``usaf_id`` values and the column ``distance_meters``.

    Returns
    -------
    isd_station, warnings : tuple of (:any:`eeweather.ISDStation`, list of str)
        A qualified weather station. ``None`` if no station meets criteria.
    """

    def _test_station(station):
        if coverage_range is None:
            return True, []
        else:
            start_date, end_date = coverage_range
            try:
                tempC, warnings = eeweather.mockable.load_isd_hourly_temp_data(
                    station, start_date, end_date, fetch_from_web
                )
            except ISDDataNotAvailableError:
                return False, []  # reject

            # TODO(philngo): also need to incorporate within-day limits
            if len(tempC) > 0:
                fraction_coverage = tempC.notnull().sum() / float(len(tempC))
                return (fraction_coverage > min_fraction_coverage), warnings
            else:
                return False, []  # reject

    def _station_warnings(station, distance_meters):
        return [
            EEWeatherWarning(
                qualified_name="eeweather.exceeds_maximum_distance",
                description=(
                    "Distance from target to weather station is greater"
                    "than the specified km."
                ),
                data={
                    "distance_meters": distance_meters,
                    "max_distance_meters": d,
                    "rank": rank,
                },
            )
            for d in distance_warnings
            if distance_meters > d
        ]

    n_stations_passed = 0
    for usaf_id, row in candidates.iterrows():
        station = ISDStation(usaf_id)
        test_result, warnings = _test_station(station)
        if test_result:
            n_stations_passed += 1
        if n_stations_passed == rank:
            if not warnings:
                warnings = []
            warnings.extend(_station_warnings(station, row.distance_meters))
            return station, warnings

    no_station_warning = EEWeatherWarning(
        qualified_name="eeweather.no_weather_station_selected",
        description=(
            "No weather station found with the specified rank and"
            " minimum fracitional coverage."
        ),
        data={"rank": rank, "min_fraction_coverage": min_fraction_coverage},
    )
    return None, [no_station_warning]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import tempfile
from eeweather.cache import KeyValueStore, get_datetime_if_exists
from datetime import datetime
import pytz

import pytest


@pytest.fixture
def s():
    return KeyValueStore("sqlite:///{}/cache.db".format(tempfile.mkdtemp()))


def test_key_value_store(s):
    # key 'a' does not exist yet
    assert s.key_exists("a") is False
    assert s.retrieve_json("a") is None
    assert s.key_updated("a") is None

    # create key 'a'
    s.save_json("a", {"b": [1, "two", 3.0]})
    assert s.key_exists("a") is True
    data = s.retrieve_json("a")
    assert len(data["b"]) == 3
    assert data["b"][0] == 1
    assert data["b"][1] == "two"
    assert data["b"][2] == 3.0
    dt1 = s.key_updated("a")
    assert dt1.date() == datetime.utcnow().date()

    # update key 'a'
    s.save_json("a", ["updated"])
    data = s.retrieve_json("a")
    assert data[0] == "updated"

    # clear key 'a' (and everything)
    s.clear()
    assert s.key_exists("a") is False


def test_key_value_store_repr(s):
    assert repr(s) == 'KeyValueStore("{}")'.format(s.url)


def test_key_value_store_clear_single_key(s):
    # clear single key
    s.save_json("a", "b")
    s.save_json("b", "c")
    assert s.key_exists("a") is True
    assert s.key_exists("b") is True
    s.clear("b")
    assert s.key_exists("a") is True
    assert s.key_exists("b") is False


def test_get_datetime_if_exists(s):
    data = None
    result = get_datetime_if_exists(data)
    assert result == None

    data = [datetime(2018, 1, 1)]
    result = get_datetime_if_exists(data)
    assert result == pytz.utc.localize(datetime(2018, 1, 1))

    data = [pytz.utc.localize(datetime(2018, 1, 1))]
    result = get_datetime_if_exists(data)
    assert result == pytz.utc.localize(datetime(2018, 1, 1))



from eeweather import get_zcta_ids, get_isd_station_usaf_ids


def test_get_zcta_ids(snapshot):
    zcta_ids = get_zcta_ids()
    snapshot.assert_match(len(zcta_ids), "n_zcta_ids")
    assert zcta_ids[0] == "00601"


def test_get_zcta_ids_by_state(snapshot):
    zcta_ids = get_zcta_ids(state="CA")
    snapshot.assert_match(len(zcta_ids), "n_zcta_ids")
    assert zcta_ids[0] == "90001"


def test_get_isd_station_usaf_ids(snapshot):
    usaf_ids = get_isd_station_usaf_ids()
    snapshot.assert_match(len(usaf_ids), "n_usaf_ids")
    assert usaf_ids[0] == "102540"


def test_get_isd_station_usaf_ids_by_state(snapshot):
    usaf_ids = get_isd_station_usaf_ids(state="IL")
    snapshot.assert_match(len(usaf_ids), "n_usaf_ids")
    assert usaf_ids[0] == "720137"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
""" Note about database tests.

    The database can be rebuilt from scratch using the command:

        $ eeweather rebuild_db

    The database can be inspected using the command (opens sqlite3 CLI):

        $ eeweather inspect_db

    These tests do not cover these main functions - instead, they test the
    content of the database. They can be thought of as integration tests.
"""
from eeweather.connections import metadata_db_connection_proxy


def test_database_tables():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select name from sqlite_master where type='table' """)
    tables = [name for (name,) in cur.fetchall()]
    assert tables == [
        "isd_station_metadata",
        "isd_file_metadata",
        "zcta_metadata",
        "iecc_climate_zone_metadata",
        "iecc_moisture_regime_metadata",
        "ba_climate_zone_metadata",
        "ca_climate_zone_metadata",
        "tmy3_station_metadata",
        "cz2010_station_metadata",
    ]


def test_isd_station_metadata_table_count(snapshot):
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from isd_station_metadata """)
    (count,) = cur.fetchone()
    snapshot.assert_match(count, "count")


def test_isd_station_metadata_table_content(snapshot):
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select * from isd_station_metadata where quality='high' limit 1 """)
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    snapshot.assert_match(data, "data")


def test_isd_file_metadata_table_count(snapshot):
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from isd_file_metadata """)
    (count,) = cur.fetchone()
    snapshot.assert_match(count, "count")


def test_isd_file_metadata_table_content(snapshot):
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select * from isd_file_metadata limit 1""")
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    snapshot.assert_match(data, "data")


def test_zcta_metadata_table_count():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from zcta_metadata """)
    (count,) = cur.fetchone()
    assert count == 33144


def test_zcta_metadata_table_content():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select * from zcta_metadata limit 1""")
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    assert data == {
        "ba_climate_zone": "Hot-Humid",
        "ca_climate_zone": None,
        "geometry": None,
        "iecc_climate_zone": "1",
        "iecc_moisture_regime": "A",
        "latitude": "18.1800455429617",
        "longitude": "-66.752178136408",
        "state": "PR",
        "zcta_id": "00601",
    }


def test_iecc_climate_zone_metadata_table_count():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from iecc_climate_zone_metadata """)
    (count,) = cur.fetchone()
    assert count == 8


def test_iecc_climate_zone_metadata_table_content():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select iecc_climate_zone from iecc_climate_zone_metadata limit 1""")
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    assert data == {"iecc_climate_zone": "1"}


def test_iecc_moisture_regime_metadata_table_count():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from iecc_moisture_regime_metadata """)
    (count,) = cur.fetchone()
    assert count == 3


def test_iecc_moisture_regime_metadata_table_content():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(
        """ select iecc_moisture_regime from iecc_moisture_regime_metadata limit 1"""
    )
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    assert data == {"iecc_moisture_regime": "A"}


def test_ba_climate_zone_metadata_table_count():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from ba_climate_zone_metadata """)
    (count,) = cur.fetchone()
    assert count == 8


def test_ba_climate_zone_metadata_table_content():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select ba_climate_zone from ba_climate_zone_metadata limit 1""")
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    assert data == {"ba_climate_zone": "Cold"}


def test_ca_climate_zone_metadata_table_count():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from ca_climate_zone_metadata """)
    (count,) = cur.fetchone()
    assert count == 16


def test_ca_climate_zone_metadata_table_content():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(
        """ select ca_climate_zone, name from ca_climate_zone_metadata limit 1"""
    )
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    assert data == {"ca_climate_zone": "CA_01", "name": "Arcata"}


def test_tmy3_station_metadata_table_count():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from tmy3_station_metadata """)
    (count,) = cur.fetchone()
    assert count == 1020


def test_tmy3_station_metadata_table_content():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select * from tmy3_station_metadata limit 1""")
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    assert data == {
        "class": "I",
        "name": "Twentynine Palms",
        "state": "CA",
        "usaf_id": "690150",
    }


def test_cz2010_station_metadata_table_count():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select count(*) from cz2010_station_metadata """)
    (count,) = cur.fetchone()
    assert count == 86


def test_cz2010_station_metadata_table_content():
    conn = metadata_db_connection_proxy.get_connection()

    cur = conn.cursor()
    cur.execute(""" select * from cz2010_station_metadata limit 1""")
    row = cur.fetchone()
    data = {desc[0]: value for value, desc in zip(row, cur.description)}
    assert data == {"usaf_id": "690150"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from datetime import datetime
import pandas as pd
import pytest
import pytz

from eeweather import (
    ISDStation,
    get_isd_station_metadata,
    get_isd_filenames,
    get_gsod_filenames,
    get_isd_file_metadata,
    fetch_isd_raw_temp_data,
    fetch_isd_hourly_temp_data,
    fetch_isd_daily_temp_data,
    fetch_gsod_raw_temp_data,
    fetch_gsod_daily_temp_data,
    fetch_tmy3_hourly_temp_data,
    fetch_cz2010_hourly_temp_data,
    get_isd_hourly_temp_data_cache_key,
    get_isd_daily_temp_data_cache_key,
    get_gsod_daily_temp_data_cache_key,
    get_tmy3_hourly_temp_data_cache_key,
    get_cz2010_hourly_temp_data_cache_key,
    cached_isd_hourly_temp_data_is_expired,
    cached_isd_daily_temp_data_is_expired,
    cached_gsod_daily_temp_data_is_expired,
    validate_isd_hourly_temp_data_cache,
    validate_isd_daily_temp_data_cache,
    validate_gsod_daily_temp_data_cache,
    validate_tmy3_hourly_temp_data_cache,
    validate_cz2010_hourly_temp_data_cache,
    serialize_isd_hourly_temp_data,
    serialize_isd_daily_temp_data,
    serialize_gsod_daily_temp_data,
    serialize_tmy3_hourly_temp_data,
    serialize_cz2010_hourly_temp_data,
    deserialize_isd_hourly_temp_data,
    deserialize_isd_daily_temp_data,
    deserialize_gsod_daily_temp_data,
    deserialize_tmy3_hourly_temp_data,
    deserialize_cz2010_hourly_temp_data,
    read_isd_hourly_temp_data_from_cache,
    read_isd_daily_temp_data_from_cache,
    read_gsod_daily_temp_data_from_cache,
    read_tmy3_hourly_temp_data_from_cache,
    read_cz2010_hourly_temp_data_from_cache,
    write_isd_hourly_temp_data_to_cache,
    write_isd_daily_temp_data_to_cache,
    write_gsod_daily_temp_data_to_cache,
    write_tmy3_hourly_temp_data_to_cache,
    write_cz2010_hourly_temp_data_to_cache,
    destroy_cached_isd_hourly_temp_data,
    destroy_cached_isd_daily_temp_data,
    destroy_cached_gsod_daily_temp_data,
    destroy_cached_tmy3_hourly_temp_data,
    destroy_cached_cz2010_hourly_temp_data,
    load_isd_hourly_temp_data_cached_proxy,
    load_isd_daily_temp_data_cached_proxy,
    load_gsod_daily_temp_data_cached_proxy,
    load_tmy3_hourly_temp_data_cached_proxy,
    load_cz2010_hourly_temp_data_cached_proxy,
    load_isd_hourly_temp_data,
    load_isd_daily_temp_data,
    load_gsod_daily_temp_data,
    load_tmy3_hourly_temp_data,
    load_cz2010_hourly_temp_data,
    load_cached_isd_hourly_temp_data,
    load_cached_isd_daily_temp_data,
    load_cached_gsod_daily_temp_data,
    load_cached_tmy3_hourly_temp_data,
    load_cached_cz2010_hourly_temp_data,
)
from eeweather.exceptions import (
    UnrecognizedUSAFIDError,
    ISDDataNotAvailableError,
    GSODDataNotAvailableError,
    TMY3DataNotAvailableError,
    CZ2010DataNotAvailableError,
    NonUTCTimezoneInfoError,
)
from eeweather.testing import (
    MockNOAAFTPConnectionProxy,
    MockKeyValueStoreProxy,
    mock_request_text_tmy3,
    mock_request_text_cz2010,
)


@pytest.fixture
def monkeypatch_noaa_ftp(monkeypatch):
    monkeypatch.setattr(
        "eeweather.connections.noaa_ftp_connection_proxy", MockNOAAFTPConnectionProxy()
    )


@pytest.fixture
def monkeypatch_tmy3_request(monkeypatch):
    monkeypatch.setattr("eeweather.mockable.request_text", mock_request_text_tmy3)


@pytest.fixture
def monkeypatch_cz2010_request(monkeypatch):
    monkeypatch.setattr("eeweather.mockable.request_text", mock_request_text_cz2010)


@pytest.fixture
def monkeypatch_key_value_store(monkeypatch):
    key_value_store_proxy = MockKeyValueStoreProxy()
    monkeypatch.setattr(
        "eeweather.connections.key_value_store_proxy", key_value_store_proxy
    )

    return key_value_store_proxy.get_store()


def test_get_isd_station_metadata():
    assert get_isd_station_metadata("722874") == {
        "ba_climate_zone": "Hot-Dry",
        "ca_climate_zone": "CA_08",
        "elevation": "+0054.6",
        "icao_code": "KCQT",
        "iecc_climate_zone": "3",
        "iecc_moisture_regime": "B",
        "latitude": "+34.024",
        "longitude": "-118.291",
        "name": "DOWNTOWN L.A./USC CAMPUS",
        "quality": "high",
        "recent_wban_id": "93134",
        "state": "CA",
        "usaf_id": "722874",
        "wban_ids": "93134",
    }


def test_isd_station_no_load_metadata():
    station = ISDStation("722880", load_metadata=False)
    assert station.usaf_id == "722880"
    assert station.iecc_climate_zone is None
    assert station.iecc_moisture_regime is None
    assert station.ba_climate_zone is None
    assert station.ca_climate_zone is None
    assert station.elevation is None
    assert station.latitude is None
    assert station.longitude is None
    assert station.coords is None
    assert station.name is None
    assert station.quality is None
    assert station.wban_ids is None
    assert station.recent_wban_id is None
    assert station.climate_zones == {}

    assert str(station) == "722880"
    assert repr(station) == "ISDStation('722880')"


def test_isd_station_no_load_metadata_invalid():
    with pytest.raises(UnrecognizedUSAFIDError):
        station = ISDStation("FAKE", load_metadata=False)


def test_isd_station_with_load_metadata():
    station = ISDStation("722880", load_metadata=True)
    assert station.usaf_id == "722880"
    assert station.iecc_climate_zone == "3"
    assert station.iecc_moisture_regime == "B"
    assert station.ba_climate_zone == "Hot-Dry"
    assert station.ca_climate_zone == "CA_09"
    assert station.elevation == 236.2
    assert station.icao_code == "KBUR"
    assert station.latitude == 34.201
    assert station.longitude == -118.358
    assert station.coords == (34.201, -118.358)
    assert station.name == "BURBANK-GLENDALE-PASA ARPT"
    assert station.quality == "high"
    assert station.wban_ids == ["23152", "99999"]
    assert station.recent_wban_id == "23152"
    assert station.climate_zones == {
        "ba_climate_zone": "Hot-Dry",
        "ca_climate_zone": "CA_09",
        "iecc_climate_zone": "3",
        "iecc_moisture_regime": "B",
    }


def test_isd_station_json():
    station = ISDStation("722880", load_metadata=True)
    assert station.json() == {
        "elevation": 236.2,
        "icao_code": "KBUR",
        "latitude": 34.201,
        "longitude": -118.358,
        "name": "BURBANK-GLENDALE-PASA ARPT",
        "quality": "high",
        "recent_wban_id": "23152",
        "wban_ids": ["23152", "99999"],
        "climate_zones": {
            "ba_climate_zone": "Hot-Dry",
            "ca_climate_zone": "CA_09",
            "iecc_climate_zone": "3",
            "iecc_moisture_regime": "B",
        },
    }


def test_isd_station_unrecognized_usaf_id():
    with pytest.raises(UnrecognizedUSAFIDError):
        station = ISDStation("FAKE", load_metadata=True)


def test_get_isd_filenames_bad_usaf_id():
    with pytest.raises(UnrecognizedUSAFIDError) as excinfo:
        get_isd_filenames("000000", 2007)
    assert excinfo.value.value == "000000"


def test_get_isd_filenames_single_year(snapshot):
    filenames = get_isd_filenames("722860", 2007)
    snapshot.assert_match(filenames, "filenames")


def test_get_isd_filenames_multiple_year(snapshot):
    filenames = get_isd_filenames("722860")
    snapshot.assert_match(filenames, "filenames")


def test_get_isd_filenames_future_year():
    filenames = get_isd_filenames("722860", 2050)
    assert filenames == ["/pub/data/noaa/2050/722860-23119-2050.gz"]


def test_get_isd_filenames_with_host():
    filenames = get_isd_filenames("722860", 2017, with_host=True)
    assert filenames == [
        "ftp://ftp.ncdc.noaa.gov/pub/data/noaa/2017/722860-23119-2017.gz"
    ]


def test_isd_station_get_isd_filenames(snapshot):
    station = ISDStation("722860")
    filenames = station.get_isd_filenames()
    snapshot.assert_match(filenames, "filenames")


def test_isd_station_get_isd_filenames_with_year(snapshot):
    station = ISDStation("722860")
    filenames = station.get_isd_filenames(2007)
    snapshot.assert_match(filenames, "filenames")


def test_isd_station_get_isd_filenames_with_host():
    station = ISDStation("722860")
    filenames = station.get_isd_filenames(2017, with_host=True)
    assert filenames == [
        "ftp://ftp.ncdc.noaa.gov/pub/data/noaa/2017/722860-23119-2017.gz"
    ]


def test_get_gsod_filenames_bad_usaf_id():
    with pytest.raises(UnrecognizedUSAFIDError) as excinfo:
        get_gsod_filenames("000000", 2007)
    assert excinfo.value.value == "000000"


def test_get_gsod_filenames_single_year(snapshot):
    filenames = get_gsod_filenames("722860", 2007)
    snapshot.assert_match(filenames, "filenames")


def test_get_gsod_filenames_multiple_year(snapshot):
    filenames = get_gsod_filenames("722860")
    snapshot.assert_match(filenames, "filenames")


def test_get_gsod_filenames_future_year():
    filenames = get_gsod_filenames("722860", 2050)
    assert filenames == ["/pub/data/gsod/2050/722860-23119-2050.op.gz"]


def test_get_gsod_filenames_with_host():
    filenames = get_gsod_filenames("722860", 2017, with_host=True)
    assert filenames == [
        "ftp://ftp.ncdc.noaa.gov/pub/data/gsod/2017/722860-23119-2017.op.gz"
    ]


def test_isd_station_get_gsod_filenames(snapshot):
    station = ISDStation("722860")
    filenames = station.get_gsod_filenames()
    snapshot.assert_match(filenames, "filenames")


def test_isd_station_get_gsod_filenames_with_year(snapshot):
    station = ISDStation("722860")
    filenames = station.get_gsod_filenames(2007)
    snapshot.assert_match(filenames, "filenames")


def test_isd_station_get_gsod_filenames_with_host():
    station = ISDStation("722860")
    filenames = station.get_gsod_filenames(2017, with_host=True)
    assert filenames == [
        "ftp://ftp.ncdc.noaa.gov/pub/data/gsod/2017/722860-23119-2017.op.gz"
    ]


def test_get_isd_file_metadata():
    assert get_isd_file_metadata("722874") == [
        {"usaf_id": "722874", "wban_id": "93134", "year": "2006"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2007"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2008"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2009"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2010"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2011"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2012"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2013"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2014"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2015"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2016"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2017"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2018"},
    ]

    with pytest.raises(UnrecognizedUSAFIDError) as excinfo:
        get_isd_file_metadata("000000")
    assert excinfo.value.value == "000000"


def test_isd_station_get_isd_file_metadata():
    station = ISDStation("722874")
    assert station.get_isd_file_metadata() == [
        {"usaf_id": "722874", "wban_id": "93134", "year": "2006"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2007"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2008"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2009"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2010"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2011"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2012"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2013"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2014"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2015"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2016"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2017"},
        {"usaf_id": "722874", "wban_id": "93134", "year": "2018"},
    ]


# fetch raw
def test_fetch_isd_raw_temp_data(monkeypatch_noaa_ftp):
    data = fetch_isd_raw_temp_data("722874", 2007)
    assert round(data.sum()) == 185945
    assert data.shape == (11094,)


def test_fetch_gsod_raw_temp_data(monkeypatch_noaa_ftp):
    data = fetch_gsod_raw_temp_data("722874", 2007)
    assert data.sum() == 6509.5
    assert data.shape == (365,)


# station fetch raw
def test_isd_station_fetch_isd_raw_temp_data(monkeypatch_noaa_ftp):
    station = ISDStation("722874")
    data = station.fetch_isd_raw_temp_data(2007)
    assert round(data.sum()) == 185945
    assert data.shape == (11094,)


def test_isd_station_fetch_gsod_raw_temp_data(monkeypatch_noaa_ftp):
    station = ISDStation("722874")
    data = station.fetch_gsod_raw_temp_data(2007)
    assert data.sum() == 6509.5
    assert data.shape == (365,)


# fetch raw invalid station
def test_fetch_isd_raw_temp_data_invalid_station():
    with pytest.raises(UnrecognizedUSAFIDError):
        fetch_isd_raw_temp_data("INVALID", 2007)


def test_fetch_gsod_raw_temp_data_invalid_station():
    with pytest.raises(UnrecognizedUSAFIDError):
        fetch_gsod_raw_temp_data("INVALID", 2007)


# fetch raw invalid year
def test_fetch_isd_raw_temp_data_invalid_year(monkeypatch_noaa_ftp):
    with pytest.raises(ISDDataNotAvailableError):
        fetch_isd_raw_temp_data("722874", 1800)


def test_fetch_gsod_raw_temp_data_invalid_year(monkeypatch_noaa_ftp):
    with pytest.raises(GSODDataNotAvailableError):
        fetch_gsod_raw_temp_data("722874", 1800)


# fetch file full of nans
def test_isd_station_fetch_isd_raw_temp_data_all_nan(monkeypatch_noaa_ftp):
    station = ISDStation("994035")
    data = station.fetch_isd_raw_temp_data(2013)
    assert round(data.sum()) == 0
    assert data.shape == (8611,)


# fetch
def test_fetch_isd_hourly_temp_data(monkeypatch_noaa_ftp):
    data = fetch_isd_hourly_temp_data("722874", 2007)
    assert data.sum() == 156160.0355
    assert data.shape == (8760,)


def test_fetch_isd_daily_temp_data(monkeypatch_noaa_ftp):
    data = fetch_isd_daily_temp_data("722874", 2007)
    assert data.sum() == 6510.002260821784
    assert data.shape == (365,)


def test_fetch_gsod_daily_temp_data(monkeypatch_noaa_ftp):
    data = fetch_gsod_daily_temp_data("722874", 2007)
    assert data.sum() == 6509.5
    assert data.shape == (365,)


def test_fetch_tmy3_hourly_temp_data(monkeypatch_tmy3_request):
    data = fetch_tmy3_hourly_temp_data("722880")
    assert data.sum() == 156194.3
    assert data.shape == (8760,)


def test_fetch_cz2010_hourly_temp_data(monkeypatch_cz2010_request):
    data = fetch_cz2010_hourly_temp_data("722880")
    assert data.sum() == 153430.90000000002
    assert data.shape == (8760,)


# station fetch
def test_isd_station_fetch_isd_hourly_temp_data(monkeypatch_noaa_ftp):
    station = ISDStation("722874")
    data = station.fetch_isd_hourly_temp_data(2007)
    assert data.sum() == 156160.0355
    assert data.shape == (8760,)


def test_isd_station_fetch_isd_daily_temp_data(monkeypatch_noaa_ftp):
    station = ISDStation("722874")
    data = station.fetch_isd_daily_temp_data(2007)
    assert data.sum() == 6510.002260821784
    assert data.shape == (365,)


def test_isd_station_fetch_gsod_daily_temp_data(monkeypatch_noaa_ftp):
    station = ISDStation("722874")
    data = station.fetch_gsod_daily_temp_data(2007)
    assert data.sum() == 6509.5
    assert data.shape == (365,)


def test_tmy3_station_hourly_temp_data(monkeypatch_tmy3_request):
    station = ISDStation("722880")
    data = station.fetch_tmy3_hourly_temp_data()
    assert data.sum() == 156194.3
    assert data.shape == (8760,)


def test_cz2010_station_hourly_temp_data(monkeypatch_cz2010_request):
    station = ISDStation("722880")
    data = station.fetch_cz2010_hourly_temp_data()
    assert data.sum() == 153430.90000000002
    assert data.shape == (8760,)


# fetch invalid station
def test_fetch_isd_hourly_temp_data_invalid():
    with pytest.raises(UnrecognizedUSAFIDError):
        fetch_isd_hourly_temp_data("INVALID", 2007)


def test_fetch_isd_daily_temp_data_invalid():
    with pytest.raises(UnrecognizedUSAFIDError):
        fetch_isd_daily_temp_data("INVALID", 2007)


def test_fetch_gsod_daily_temp_data_invalid():
    with pytest.raises(UnrecognizedUSAFIDError):
        fetch_gsod_daily_temp_data("INVALID", 2007)


def test_fetch_tmy3_hourly_temp_data_invalid():
    with pytest.raises(TMY3DataNotAvailableError):
        fetch_tmy3_hourly_temp_data("INVALID")


def test_fetch_cz2010_hourly_temp_data_invalid():
    with pytest.raises(CZ2010DataNotAvailableError):
        fetch_cz2010_hourly_temp_data("INVALID")


def test_fetch_tmy3_hourly_temp_data_not_in_tmy3_list(monkeypatch_noaa_ftp):
    data = fetch_isd_hourly_temp_data("722874", 2007)
    assert data.sum() == 156160.0355
    assert data.shape == (8760,)
    with pytest.raises(TMY3DataNotAvailableError):
        fetch_tmy3_hourly_temp_data("722874")


def test_fetch_cz2010_hourly_temp_data_not_in_cz2010_list(monkeypatch_cz2010_request):
    data = fetch_cz2010_hourly_temp_data("722880")
    assert data.sum() == 153430.90000000002
    assert data.shape == (8760,)
    with pytest.raises(CZ2010DataNotAvailableError):
        fetch_cz2010_hourly_temp_data("725340")


# get cache key
def test_get_isd_hourly_temp_data_cache_key():
    assert (
        get_isd_hourly_temp_data_cache_key("722874", 2007) == "isd-hourly-722874-2007"
    )


def test_get_isd_daily_temp_data_cache_key():
    assert get_isd_daily_temp_data_cache_key("722874", 2007) == "isd-daily-722874-2007"


def test_get_gsod_daily_temp_data_cache_key():
    assert (
        get_gsod_daily_temp_data_cache_key("722874", 2007) == "gsod-daily-722874-2007"
    )


def test_get_tmy3_hourly_temp_data_cache_key():
    assert get_tmy3_hourly_temp_data_cache_key("722880") == "tmy3-hourly-722880"


def test_get_cz2010_hourly_temp_data_cache_key():
    assert get_cz2010_hourly_temp_data_cache_key("722880") == "cz2010-hourly-722880"


# station get cache key
def test_isd_station_get_isd_hourly_temp_data_cache_key():
    station = ISDStation("722874")
    assert station.get_isd_hourly_temp_data_cache_key(2007) == "isd-hourly-722874-2007"


def test_isd_station_get_isd_daily_temp_data_cache_key():
    station = ISDStation("722874")
    assert station.get_isd_daily_temp_data_cache_key(2007) == "isd-daily-722874-2007"


def test_isd_station_get_gsod_daily_temp_data_cache_key():
    station = ISDStation("722874")
    assert station.get_gsod_daily_temp_data_cache_key(2007) == "gsod-daily-722874-2007"


def test_tmy3_station_get_isd_hourly_temp_data_cache_key():
    station = ISDStation("722880")
    assert station.get_tmy3_hourly_temp_data_cache_key() == "tmy3-hourly-722880"


def test_cz2010_station_get_isd_hourly_temp_data_cache_key():
    station = ISDStation("722880")
    assert station.get_cz2010_hourly_temp_data_cache_key() == "cz2010-hourly-722880"


# cache expired empty
def test_cached_isd_hourly_temp_data_is_expired_empty(monkeypatch_key_value_store):
    assert cached_isd_hourly_temp_data_is_expired("722874", 2007) is True


def test_cached_isd_daily_temp_data_is_expired_empty(monkeypatch_key_value_store):
    assert cached_isd_daily_temp_data_is_expired("722874", 2007) is True


def test_cached_gsod_daily_temp_data_is_expired_empty(monkeypatch_key_value_store):
    assert cached_gsod_daily_temp_data_is_expired("722874", 2007) is True


# station cache expired empty
def test_isd_station_cached_isd_hourly_temp_data_is_expired_empty(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    assert station.cached_isd_hourly_temp_data_is_expired(2007) is True


def test_isd_station_cached_isd_daily_temp_data_is_expired_empty(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    assert station.cached_isd_daily_temp_data_is_expired(2007) is True


def test_isd_station_cached_gsod_daily_temp_data_is_expired_empty(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    assert station.cached_gsod_daily_temp_data_is_expired(2007) is True


# cache expired false
def test_cached_isd_hourly_temp_data_is_expired_false(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_isd_hourly_temp_data_cached_proxy("722874", 2007)
    assert cached_isd_hourly_temp_data_is_expired("722874", 2007) is False


def test_cached_isd_daily_temp_data_is_expired_false(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_isd_daily_temp_data_cached_proxy("722874", 2007)
    assert cached_isd_daily_temp_data_is_expired("722874", 2007) is False


def test_cached_gsod_daily_temp_data_is_expired_false(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_gsod_daily_temp_data_cached_proxy("722874", 2007)
    assert cached_gsod_daily_temp_data_is_expired("722874", 2007) is False


# cache expired true
def test_cached_isd_hourly_temp_data_is_expired_true(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_isd_hourly_temp_data_cached_proxy("722874", 2007)

    # manually expire key value item
    key = get_isd_hourly_temp_data_cache_key("722874", 2007)
    store = monkeypatch_key_value_store
    store.items.update().where(store.items.c.key == key).values(
        updated=pytz.UTC.localize(datetime(2007, 3, 3))
    ).execute()

    assert cached_isd_hourly_temp_data_is_expired("722874", 2007) is True


def test_cached_isd_daily_temp_data_is_expired_true(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_isd_daily_temp_data_cached_proxy("722874", 2007)

    # manually expire key value item
    key = get_isd_daily_temp_data_cache_key("722874", 2007)
    store = monkeypatch_key_value_store
    store.items.update().where(store.items.c.key == key).values(
        updated=pytz.UTC.localize(datetime(2007, 3, 3))
    ).execute()

    assert cached_isd_daily_temp_data_is_expired("722874", 2007) is True


def test_cached_gsod_daily_temp_data_is_expired_true(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_gsod_daily_temp_data_cached_proxy("722874", 2007)

    # manually expire key value item
    key = get_gsod_daily_temp_data_cache_key("722874", 2007)
    store = monkeypatch_key_value_store
    store.items.update().where(store.items.c.key == key).values(
        updated=pytz.UTC.localize(datetime(2007, 3, 3))
    ).execute()

    assert cached_gsod_daily_temp_data_is_expired("722874", 2007) is True


# validate cache empty
def test_validate_isd_hourly_temp_data_cache_empty(monkeypatch_key_value_store):
    assert validate_isd_hourly_temp_data_cache("722874", 2007) is False


def test_validate_isd_daily_temp_data_cache_empty(monkeypatch_key_value_store):
    assert validate_isd_daily_temp_data_cache("722874", 2007) is False


def test_validate_gsod_daily_temp_data_cache_empty(monkeypatch_key_value_store):
    assert validate_gsod_daily_temp_data_cache("722874", 2007) is False


def test_validate_tmy3_hourly_temp_data_cache_empty(monkeypatch_key_value_store):
    assert validate_tmy3_hourly_temp_data_cache("722880") is False


def test_validate_cz2010_hourly_temp_data_cache_empty(monkeypatch_key_value_store):
    assert validate_cz2010_hourly_temp_data_cache("722880") is False


# station validate cache empty
def test_isd_station_validate_isd_hourly_temp_data_cache_empty(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    assert station.validate_isd_hourly_temp_data_cache(2007) is False


def test_isd_station_validate_isd_daily_temp_data_cache_empty(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    assert station.validate_isd_daily_temp_data_cache(2007) is False


def test_isd_station_validate_gsod_daily_temp_data_cache_empty(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    assert station.validate_gsod_daily_temp_data_cache(2007) is False


def test_isd_station_validate_tmy3_hourly_temp_data_cache_empty(
    monkeypatch_key_value_store
):
    station = ISDStation("722880")
    assert station.validate_tmy3_hourly_temp_data_cache() is False


def test_isd_station_validate_cz2010_hourly_temp_data_cache_empty(
    monkeypatch_key_value_store
):
    station = ISDStation("722880")
    assert station.validate_cz2010_hourly_temp_data_cache() is False


# error on non-existent when relying on cache
def test_raise_on_missing_isd_hourly_temp_data_cache_data_no_web_fetch(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    with pytest.raises(ISDDataNotAvailableError):
        load_isd_hourly_temp_data_cached_proxy("722874", 1907, fetch_from_web=False)


def test_raise_on_missing_isd_daily_temp_data_cache_data_no_web_fetch(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    with pytest.raises(ISDDataNotAvailableError):
        load_isd_daily_temp_data_cached_proxy("722874", 1907, fetch_from_web=False)


def test_raise_on_missing_gsod_daily_temp_data_cache_data_no_web_fetch(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    with pytest.raises(GSODDataNotAvailableError):
        load_gsod_daily_temp_data_cached_proxy("722874", 1907, fetch_from_web=False)


def test_raise_on_missing_tmy3_hourly_temp_data_cache_data_no_web_fetch(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    with pytest.raises(TMY3DataNotAvailableError):
        load_tmy3_hourly_temp_data_cached_proxy("722874", fetch_from_web=False)


def test_raise_on_missing_cz2010_hourly_temp_data_cache_data_no_web_fetch(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    with pytest.raises(CZ2010DataNotAvailableError):
        load_cz2010_hourly_temp_data_cached_proxy("722874", fetch_from_web=False)


# validate updated recently
def test_validate_isd_hourly_temp_data_cache_updated_recently(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_isd_hourly_temp_data_cached_proxy("722874", 2007)
    assert validate_isd_hourly_temp_data_cache("722874", 2007) is True


def test_validate_isd_daily_temp_data_cache_updated_recently(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_isd_daily_temp_data_cached_proxy("722874", 2007)
    assert validate_isd_daily_temp_data_cache("722874", 2007) is True


def test_validate_gsod_daily_temp_data_cache_updated_recently(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_gsod_daily_temp_data_cached_proxy("722874", 2007)
    assert validate_gsod_daily_temp_data_cache("722874", 2007) is True


def test_validate_tmy3_hourly_temp_data_cache_updated_recently(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):
    load_tmy3_hourly_temp_data_cached_proxy("722880")
    assert validate_tmy3_hourly_temp_data_cache("722880") is True


def test_validate_cz2010_hourly_temp_data_cache_updated_recently(
    monkeypatch_cz2010_request, monkeypatch_key_value_store
):
    load_cz2010_hourly_temp_data_cached_proxy("722880")
    assert validate_cz2010_hourly_temp_data_cache("722880") is True


# validate expired
def test_validate_isd_hourly_temp_data_cache_expired(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_isd_hourly_temp_data_cached_proxy("722874", 2007)

    # manually expire key value item
    key = get_isd_hourly_temp_data_cache_key("722874", 2007)
    store = monkeypatch_key_value_store
    store.items.update().where(store.items.c.key == key).values(
        updated=pytz.UTC.localize(datetime(2007, 3, 3))
    ).execute()

    assert validate_isd_hourly_temp_data_cache("722874", 2007) is False


def test_validate_isd_daily_temp_data_cache_expired(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_isd_daily_temp_data_cached_proxy("722874", 2007)

    # manually expire key value item
    key = get_isd_daily_temp_data_cache_key("722874", 2007)
    store = monkeypatch_key_value_store
    store.items.update().where(store.items.c.key == key).values(
        updated=pytz.UTC.localize(datetime(2007, 3, 3))
    ).execute()

    assert validate_isd_daily_temp_data_cache("722874", 2007) is False


def test_validate_gsod_daily_temp_data_cache_expired(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    load_gsod_daily_temp_data_cached_proxy("722874", 2007)

    # manually expire key value item
    key = get_gsod_daily_temp_data_cache_key("722874", 2007)
    store = monkeypatch_key_value_store
    store.items.update().where(store.items.c.key == key).values(
        updated=pytz.UTC.localize(datetime(2007, 3, 3))
    ).execute()

    assert validate_gsod_daily_temp_data_cache("722874", 2007) is False


# serialize
def test_serialize_isd_hourly_temp_data():
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert serialize_isd_hourly_temp_data(ts) == [["2017010100", 1]]


def test_serialize_isd_daily_temp_data():
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert serialize_isd_daily_temp_data(ts) == [["20170101", 1]]


def test_serialize_gsod_daily_temp_data():
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert serialize_gsod_daily_temp_data(ts) == [["20170101", 1]]


def test_serialize_tmy3_hourly_temp_data():
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert serialize_tmy3_hourly_temp_data(ts) == [["2017010100", 1]]


def test_serialize_cz2010_hourly_temp_data():
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert serialize_cz2010_hourly_temp_data(ts) == [["2017010100", 1]]


# station serialize
def test_isd_station_serialize_isd_hourly_temp_data():
    station = ISDStation("722874")
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert station.serialize_isd_hourly_temp_data(ts) == [["2017010100", 1]]


def test_isd_station_serialize_isd_daily_temp_data():
    station = ISDStation("722874")
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert station.serialize_isd_daily_temp_data(ts) == [["20170101", 1]]


def test_isd_station_serialize_gsod_daily_temp_data():
    station = ISDStation("722874")
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert station.serialize_gsod_daily_temp_data(ts) == [["20170101", 1]]


def test_isd_station_serialize_tmy3_hourly_temp_data():
    station = ISDStation("722880")
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert station.serialize_tmy3_hourly_temp_data(ts) == [["2017010100", 1]]


def test_isd_station_serialize_cz2010_hourly_temp_data():
    station = ISDStation("722880")
    ts = pd.Series([1], index=[pytz.UTC.localize(datetime(2017, 1, 1))])
    assert station.serialize_cz2010_hourly_temp_data(ts) == [["2017010100", 1]]


# deserialize
def test_deserialize_isd_hourly_temp_data():
    ts = deserialize_isd_hourly_temp_data([["2017010100", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "H"


def test_deserialize_isd_daily_temp_data():
    ts = deserialize_isd_daily_temp_data([["20170101", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "D"


def test_deserialize_gsod_daily_temp_data():
    ts = deserialize_gsod_daily_temp_data([["20170101", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "D"


def test_deserialize_tmy3_hourly_temp_data():
    ts = deserialize_tmy3_hourly_temp_data([["2017010100", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "H"


def test_deserialize_cz2010_hourly_temp_data():
    ts = deserialize_cz2010_hourly_temp_data([["2017010100", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "H"


# station deserialize
def test_isd_station_deserialize_isd_hourly_temp_data():
    station = ISDStation("722874")
    ts = station.deserialize_isd_hourly_temp_data([["2017010100", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "H"


def test_isd_station_deserialize_isd_daily_temp_data():
    station = ISDStation("722874")
    ts = station.deserialize_isd_daily_temp_data([["20170101", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "D"


def test_isd_station_deserialize_gsod_daily_temp_data():
    station = ISDStation("722874")
    ts = station.deserialize_gsod_daily_temp_data([["20170101", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "D"


def test_isd_station_deserialize_tmy3_hourly_temp_data():
    station = ISDStation("722880")
    ts = station.deserialize_tmy3_hourly_temp_data([["2017010100", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "H"


def test_isd_station_deserialize_cz2010_hourly_temp_data():
    station = ISDStation("722880")
    ts = station.deserialize_cz2010_hourly_temp_data([["2017010100", 1]])
    assert ts.sum() == 1
    assert ts.index.freq.name == "H"


# write read destroy
def test_write_read_destroy_isd_hourly_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    store = monkeypatch_key_value_store
    key = get_isd_hourly_temp_data_cache_key("123456", 1990)
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    write_isd_hourly_temp_data_to_cache("123456", 1990, ts1)
    assert store.key_exists(key) is True

    ts2 = read_isd_hourly_temp_data_from_cache("123456", 1990)
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    destroy_cached_isd_hourly_temp_data("123456", 1990)
    assert store.key_exists(key) is False


def test_write_read_destroy_isd_daily_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    store = monkeypatch_key_value_store
    key = get_isd_daily_temp_data_cache_key("123456", 1990)
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    write_isd_daily_temp_data_to_cache("123456", 1990, ts1)
    assert store.key_exists(key) is True

    ts2 = read_isd_daily_temp_data_from_cache("123456", 1990)
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    destroy_cached_isd_daily_temp_data("123456", 1990)
    assert store.key_exists(key) is False


def test_write_read_destroy_gsod_daily_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    store = monkeypatch_key_value_store
    key = get_gsod_daily_temp_data_cache_key("123456", 1990)
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    write_gsod_daily_temp_data_to_cache("123456", 1990, ts1)
    assert store.key_exists(key) is True

    ts2 = read_gsod_daily_temp_data_from_cache("123456", 1990)
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    destroy_cached_gsod_daily_temp_data("123456", 1990)
    assert store.key_exists(key) is False


def test_write_read_destroy_tmy3_hourly_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    store = monkeypatch_key_value_store
    key = get_tmy3_hourly_temp_data_cache_key("123456")
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    write_tmy3_hourly_temp_data_to_cache("123456", ts1)
    assert store.key_exists(key) is True

    ts2 = read_tmy3_hourly_temp_data_from_cache("123456")
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    destroy_cached_tmy3_hourly_temp_data("123456")
    assert store.key_exists(key) is False


def test_write_read_destroy_cz2010_hourly_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    store = monkeypatch_key_value_store
    key = get_cz2010_hourly_temp_data_cache_key("123456")
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    write_cz2010_hourly_temp_data_to_cache("123456", ts1)
    assert store.key_exists(key) is True

    ts2 = read_cz2010_hourly_temp_data_from_cache("123456")
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    destroy_cached_cz2010_hourly_temp_data("123456")
    assert store.key_exists(key) is False


# station write read destroy
def test_isd_station_write_read_destroy_isd_hourly_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    store = monkeypatch_key_value_store
    key = station.get_isd_hourly_temp_data_cache_key(1990)
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    station.write_isd_hourly_temp_data_to_cache(1990, ts1)
    assert store.key_exists(key) is True

    ts2 = station.read_isd_hourly_temp_data_from_cache(1990)
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    station.destroy_cached_isd_hourly_temp_data(1990)
    assert store.key_exists(key) is False


def test_isd_station_write_read_destroy_isd_daily_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    store = monkeypatch_key_value_store
    key = station.get_isd_daily_temp_data_cache_key(1990)
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    station.write_isd_daily_temp_data_to_cache(1990, ts1)
    assert store.key_exists(key) is True

    ts2 = station.read_isd_daily_temp_data_from_cache(1990)
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    station.destroy_cached_isd_daily_temp_data(1990)
    assert store.key_exists(key) is False


def test_isd_station_write_read_destroy_gsod_daily_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    station = ISDStation("722874")
    store = monkeypatch_key_value_store
    key = station.get_gsod_daily_temp_data_cache_key(1990)
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    station.write_gsod_daily_temp_data_to_cache(1990, ts1)
    assert store.key_exists(key) is True

    ts2 = station.read_gsod_daily_temp_data_from_cache(1990)
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    station.destroy_cached_gsod_daily_temp_data(1990)
    assert store.key_exists(key) is False


def test_isd_station_write_read_destroy_tmy3_hourly_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    station = ISDStation("722880")
    store = monkeypatch_key_value_store
    key = station.get_tmy3_hourly_temp_data_cache_key()
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    station.write_tmy3_hourly_temp_data_to_cache(ts1)
    assert store.key_exists(key) is True

    ts2 = station.read_tmy3_hourly_temp_data_from_cache()
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    station.destroy_cached_tmy3_hourly_temp_data()
    assert store.key_exists(key) is False


def test_isd_station_write_read_destroy_cz2010_hourly_temp_data_to_from_cache(
    monkeypatch_key_value_store
):
    station = ISDStation("722880")
    store = monkeypatch_key_value_store
    key = station.get_cz2010_hourly_temp_data_cache_key()
    assert store.key_exists(key) is False

    ts1 = pd.Series([1], index=[pytz.UTC.localize(datetime(1990, 1, 1))])
    station.write_cz2010_hourly_temp_data_to_cache(ts1)
    assert store.key_exists(key) is True

    ts2 = station.read_cz2010_hourly_temp_data_from_cache()
    assert store.key_exists(key) is True
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape

    station.destroy_cached_cz2010_hourly_temp_data()
    assert store.key_exists(key) is False


# load cached proxy
def test_load_isd_hourly_temp_data_cached_proxy(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = load_isd_hourly_temp_data_cached_proxy("722874", 2007)
    ts2 = load_isd_hourly_temp_data_cached_proxy("722874", 2007)
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


def test_load_isd_daily_temp_data_cached_proxy(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = load_isd_daily_temp_data_cached_proxy("722874", 2007)
    ts2 = load_isd_daily_temp_data_cached_proxy("722874", 2007)
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


def test_load_gsod_daily_temp_data_cached_proxy(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = load_gsod_daily_temp_data_cached_proxy("722874", 2007)
    ts2 = load_gsod_daily_temp_data_cached_proxy("722874", 2007)
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


def test_load_tmy3_hourly_temp_data_cached_proxy(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = load_tmy3_hourly_temp_data_cached_proxy("722880", 2007)
    ts2 = load_tmy3_hourly_temp_data_cached_proxy("722880", 2007)
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


def test_load_cz2010_hourly_temp_data_cached_proxy(
    monkeypatch_cz2010_request, monkeypatch_key_value_store
):

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = load_cz2010_hourly_temp_data_cached_proxy("722880", 2007)
    ts2 = load_cz2010_hourly_temp_data_cached_proxy("722880", 2007)
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


# station load cached proxy
def test_isd_station_load_isd_hourly_temp_data_cached_proxy(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    station = ISDStation("722874")

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = station.load_isd_hourly_temp_data_cached_proxy(2007)
    ts2 = station.load_isd_hourly_temp_data_cached_proxy(2007)
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


def test_isd_station_load_isd_daily_temp_data_cached_proxy(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    station = ISDStation("722874")

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = station.load_isd_daily_temp_data_cached_proxy(2007)
    ts2 = station.load_isd_daily_temp_data_cached_proxy(2007)
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


def test_isd_station_load_gsod_daily_temp_data_cached_proxy(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    station = ISDStation("722874")

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = station.load_gsod_daily_temp_data_cached_proxy(2007)
    ts2 = station.load_gsod_daily_temp_data_cached_proxy(2007)
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


def test_isd_station_load_tmy3_hourly_temp_data_cached_proxy(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):
    station = ISDStation("722880")

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = station.load_tmy3_hourly_temp_data_cached_proxy()
    ts2 = station.load_tmy3_hourly_temp_data_cached_proxy()
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


def test_isd_station_load_cz2010_hourly_temp_data_cached_proxy(
    monkeypatch_cz2010_request, monkeypatch_key_value_store
):
    station = ISDStation("722880")

    # doesn't yet guarantee that all code paths are taken,
    # except that coverage picks it up either here or elsewhere
    ts1 = station.load_cz2010_hourly_temp_data_cached_proxy()
    ts2 = station.load_cz2010_hourly_temp_data_cached_proxy()
    assert int(ts1.sum()) == int(ts2.sum())
    assert ts1.shape == ts2.shape


# load data between dates
def test_load_isd_hourly_temp_data(monkeypatch_noaa_ftp, monkeypatch_key_value_store):

    start = datetime(2006, 1, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts, warnings = load_isd_hourly_temp_data("722874", start, end)
    assert ts.index[0] == start
    assert pd.isnull(ts[0])
    assert ts.index[-1] == end
    assert pd.notnull(ts[-1])


def test_load_isd_hourly_temp_data_non_normalized_dates(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    start = datetime(2006, 1, 3, 11, 12, 13, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, 12, 13, 14, tzinfo=pytz.UTC)
    ts, warnings = load_isd_hourly_temp_data("722874", start, end)
    assert ts.index[0] == datetime(2006, 1, 3, 12, tzinfo=pytz.UTC)
    assert pd.isnull(ts[0])
    assert ts.index[-1] == datetime(2007, 4, 3, 12, tzinfo=pytz.UTC)
    assert pd.notnull(ts[-1])


def test_load_isd_daily_temp_data(monkeypatch_noaa_ftp, monkeypatch_key_value_store):

    start = datetime(2006, 1, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts = load_isd_daily_temp_data("722874", start, end)
    assert ts.index[0] == start
    assert pd.isnull(ts[0])
    assert ts.index[-1] == end
    assert pd.notnull(ts[-1])


def test_load_isd_daily_temp_data_non_normalized_dates(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    start = datetime(2006, 1, 3, 11, 12, 13, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, 12, 13, 14, tzinfo=pytz.UTC)
    ts = load_isd_daily_temp_data("722874", start, end)
    assert ts.index[0] == datetime(2006, 1, 4, tzinfo=pytz.UTC)
    assert pd.isnull(ts[0])
    assert ts.index[-1] == datetime(2007, 4, 3, tzinfo=pytz.UTC)
    assert pd.notnull(ts[-1])


def test_load_gsod_daily_temp_data(monkeypatch_noaa_ftp, monkeypatch_key_value_store):

    start = datetime(2006, 1, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts = load_gsod_daily_temp_data("722874", start, end)
    assert ts.index[0] == start
    assert pd.isnull(ts[0])
    assert ts.index[-1] == end
    assert pd.notnull(ts[-1])


def test_load_gsod_daily_temp_data_non_normalized_dates(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    start = datetime(2006, 1, 3, 11, 12, 13, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, 12, 13, 14, tzinfo=pytz.UTC)
    ts = load_gsod_daily_temp_data("722874", start, end)
    assert ts.index[0] == datetime(2006, 1, 4, tzinfo=pytz.UTC)
    assert pd.isnull(ts[0])
    assert ts.index[-1] == datetime(2007, 4, 3, tzinfo=pytz.UTC)
    assert pd.notnull(ts[-1])


def test_load_tmy3_hourly_temp_data(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    start = datetime(2006, 1, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts = load_tmy3_hourly_temp_data("722880", start, end)
    assert ts.index[0] == start
    assert pd.notnull(ts[0])
    assert ts.index[-1] == end
    assert pd.notnull(ts[-1])


def test_load_cz2010_hourly_temp_data(
    monkeypatch_cz2010_request, monkeypatch_key_value_store
):

    start = datetime(2006, 1, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts = load_cz2010_hourly_temp_data("722880", start, end)
    assert ts.index[0] == start
    assert pd.notnull(ts[1])
    assert ts.index[-1] == end
    assert pd.notnull(ts[-1])


# station load data between dates
def test_isd_station_load_isd_hourly_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    station = ISDStation("722874")
    start = datetime(2007, 3, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts, warnings = station.load_isd_hourly_temp_data(start, end)
    assert ts.index[0] == start
    assert ts.index[-1] == end


def test_isd_station_load_isd_daily_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    station = ISDStation("722874")
    start = datetime(2007, 3, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts = station.load_isd_daily_temp_data(start, end)
    assert ts.index[0] == start
    assert ts.index[-1] == end


def test_isd_station_load_gsod_daily_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    station = ISDStation("722874")
    start = datetime(2007, 3, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts = station.load_gsod_daily_temp_data(start, end)
    assert ts.index[0] == start
    assert ts.index[-1] == end


def test_isd_station_load_tmy3_hourly_temp_data(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    station = ISDStation("722880")
    start = datetime(2007, 3, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts = station.load_tmy3_hourly_temp_data(start, end)
    assert ts.index[0] == start
    assert ts.index[-1] == end


def test_isd_station_load_cz2010_hourly_temp_data(
    monkeypatch_cz2010_request, monkeypatch_key_value_store
):

    station = ISDStation("722880")
    start = datetime(2007, 3, 3, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 3, tzinfo=pytz.UTC)
    ts = station.load_cz2010_hourly_temp_data(start, end)
    assert ts.index[0] == start
    assert ts.index[-1] == end


# load cached
def test_load_cached_isd_hourly_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    ts = load_cached_isd_hourly_temp_data("722874")
    assert ts is None

    # load data
    ts = load_isd_hourly_temp_data_cached_proxy("722874", 2007)
    assert int(ts.sum()) == 156160
    assert ts.shape == (8760,)

    ts = load_cached_isd_hourly_temp_data("722874")
    assert int(ts.sum()) == 156160
    assert ts.shape == (8760,)


def test_load_cached_isd_daily_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    ts = load_cached_isd_daily_temp_data("722874")
    assert ts is None

    # load data
    ts = load_isd_daily_temp_data_cached_proxy("722874", 2007)
    assert int(ts.sum()) == 6510
    assert ts.shape == (365,)

    ts = load_cached_isd_daily_temp_data("722874")
    assert int(ts.sum()) == 6510
    assert ts.shape == (365,)


def test_load_cached_gsod_daily_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):

    ts = load_cached_gsod_daily_temp_data("722874")
    assert ts is None

    # load data
    ts = load_gsod_daily_temp_data_cached_proxy("722874", 2007)
    assert int(ts.sum()) == 6509
    assert ts.shape == (365,)

    ts = load_cached_gsod_daily_temp_data("722874")
    assert int(ts.sum()) == 6509
    assert ts.shape == (365,)


def test_load_cached_tmy3_hourly_temp_data(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    ts = load_cached_tmy3_hourly_temp_data("722880")
    assert ts is None

    # load data
    ts = load_tmy3_hourly_temp_data_cached_proxy("722880")
    assert int(ts.sum()) == 156194
    assert ts.shape == (8760,)

    ts = load_cached_tmy3_hourly_temp_data("722880")
    assert int(ts.sum()) == 156194
    assert ts.shape == (8760,)


def test_load_cached_cz2010_hourly_temp_data(
    monkeypatch_cz2010_request, monkeypatch_key_value_store
):

    ts = load_cached_cz2010_hourly_temp_data("722880")
    assert ts is None

    # load data
    ts = load_cz2010_hourly_temp_data_cached_proxy("722880")
    assert int(ts.sum()) == 153430
    assert ts.shape == (8760,)

    ts = load_cached_cz2010_hourly_temp_data("722880")
    assert int(ts.sum()) == 153430
    assert ts.shape == (8760,)


# station load cached
def test_isd_station_load_cached_isd_hourly_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    station = ISDStation("722874")

    ts = station.load_cached_isd_hourly_temp_data()
    assert ts is None

    # load data
    ts = station.load_isd_hourly_temp_data_cached_proxy(2007)
    assert int(ts.sum()) == 156160
    assert ts.shape == (8760,)

    ts = station.load_cached_isd_hourly_temp_data()
    assert int(ts.sum()) == 156160
    assert ts.shape == (8760,)


def test_isd_station_load_cached_isd_daily_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    station = ISDStation("722874")

    ts = station.load_cached_isd_daily_temp_data()
    assert ts is None

    # load data
    ts = station.load_isd_daily_temp_data_cached_proxy(2007)
    assert int(ts.sum()) == 6510
    assert ts.shape == (365,)

    ts = station.load_cached_isd_daily_temp_data()
    assert int(ts.sum()) == 6510
    assert ts.shape == (365,)


def test_isd_station_load_cached_gsod_daily_temp_data(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    station = ISDStation("722874")

    ts = station.load_cached_gsod_daily_temp_data()
    assert ts is None

    # load data
    ts = station.load_gsod_daily_temp_data_cached_proxy(2007)
    assert int(ts.sum()) == 6509
    assert ts.shape == (365,)

    ts = station.load_cached_gsod_daily_temp_data()
    assert int(ts.sum()) == 6509
    assert ts.shape == (365,)


def test_isd_station_load_cached_tmy3_hourly_temp_data(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):
    station = ISDStation("722880")

    ts = station.load_cached_tmy3_hourly_temp_data()
    assert ts is None

    # load data
    ts = station.load_tmy3_hourly_temp_data_cached_proxy()
    assert int(ts.sum()) == 156194
    assert ts.shape == (8760,)

    ts = station.load_cached_tmy3_hourly_temp_data()
    assert int(ts.sum()) == 156194
    assert ts.shape == (8760,)


def test_isd_station_load_cached_cz2010_hourly_temp_data(
    monkeypatch_cz2010_request, monkeypatch_key_value_store
):
    station = ISDStation("722880")

    ts = station.load_cached_cz2010_hourly_temp_data()
    assert ts is None

    # load data
    ts = station.load_cz2010_hourly_temp_data_cached_proxy()
    assert int(ts.sum()) == 153430
    assert ts.shape == (8760,)

    ts = station.load_cached_cz2010_hourly_temp_data()
    assert int(ts.sum()) == 153430
    assert ts.shape == (8760,)


# test slicing of normalized data
def test_load_correctly_sliced_tmy3_hourly_temp_data(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    start = datetime(2015, 2, 15, tzinfo=pytz.UTC)
    end = datetime(2016, 8, 12, tzinfo=pytz.UTC)

    ts = load_tmy3_hourly_temp_data("722880", start, end)
    ts_orig = fetch_tmy3_hourly_temp_data("722880")

    for i in ts.index:
        # leap day is null
        if i.month == 2 and i.day == 29:
            assert pd.isnull(ts[i])
        else:
            assert ts[i] == ts_orig[i.replace(year=1900)]


def test_load_correctly_sliced_cz2010_hourly_temp_data(
    monkeypatch_cz2010_request, monkeypatch_key_value_store
):

    start = datetime(2015, 2, 15, tzinfo=pytz.UTC)
    end = datetime(2016, 8, 12, tzinfo=pytz.UTC)

    ts = load_cz2010_hourly_temp_data("722880", start, end)
    ts_orig = fetch_cz2010_hourly_temp_data("722880")

    for i in ts.index:
        # leap day is null
        if i.month == 2 and i.day == 29:
            assert pd.isnull(ts[i])
        else:
            assert ts[i] == ts_orig[i.replace(year=1900)]


def test_isd_station_load_isd_hourly_temp_data_tz_exception(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    station = ISDStation("722880")
    start = datetime(2007, 4, 10)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_isd_hourly_temp_data(start, end)

    start = datetime(2007, 4, 10, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_isd_hourly_temp_data(start, end)


def test_isd_station_load_isd_daily_temp_data_tz_exception(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    station = ISDStation("722880")
    start = datetime(2007, 4, 10)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_isd_daily_temp_data(start, end)

    start = datetime(2007, 4, 10, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_isd_daily_temp_data(start, end)


def test_isd_station_load_gsod_daily_temp_data_tz_exception(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    station = ISDStation("722880")
    start = datetime(2007, 4, 10)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_gsod_daily_temp_data(start, end)

    start = datetime(2007, 4, 10, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_gsod_daily_temp_data(start, end)


def test_isd_station_load_tmy3_hourly_temp_data_tz_exception(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    station = ISDStation("722880")
    start = datetime(2007, 4, 10)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_tmy3_hourly_temp_data(start, end)

    start = datetime(2007, 4, 10, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_tmy3_hourly_temp_data(start, end)


def test_isd_station_load_cz2010_hourly_temp_data_tz_exception(
    monkeypatch_tmy3_request, monkeypatch_key_value_store
):

    station = ISDStation("722880")
    start = datetime(2007, 4, 10)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_cz2010_hourly_temp_data(start, end)

    start = datetime(2007, 4, 10, tzinfo=pytz.UTC)
    end = datetime(2007, 4, 12)
    with pytest.raises(NonUTCTimezoneInfoError):
        ts = station.load_cz2010_hourly_temp_data(start, end)


def test_isd_station_metadata_null_elevation():
    usaf_id = "724953"
    metadata = get_isd_station_metadata(usaf_id)
    assert metadata["elevation"] is None
    isd_station = ISDStation(usaf_id)
    assert isd_station.elevation is None


def test_load_isd_hourly_temp_data_missing_years(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    usaf_id = "722874"
    start = datetime(2005, 1, 1, tzinfo=pytz.UTC)
    end = datetime(2007, 12, 31, tzinfo=pytz.UTC)
    with pytest.raises(ISDDataNotAvailableError):
        ts = load_isd_hourly_temp_data(usaf_id, start, end, error_on_missing_years=True)
    ts, warnings = load_isd_hourly_temp_data(
        usaf_id, start, end, error_on_missing_years=False
    )
    assert ts.index[0] == start
    assert pd.isnull(ts[0])
    assert ts.index[-1] == end
    assert pd.notnull(ts[-1])


def test_isd_station_load_isd_hourly_temp_data_missing_years(
    monkeypatch_noaa_ftp, monkeypatch_key_value_store
):
    usaf_id = "722874"
    start = datetime(2005, 1, 1, tzinfo=pytz.UTC)
    end = datetime(2007, 12, 31, tzinfo=pytz.UTC)
    isd_station = ISDStation(usaf_id)
    with pytest.raises(ISDDataNotAvailableError):
        isd_station.load_isd_hourly_temp_data(start, end, error_on_missing_years=True)
    ts, warnings = isd_station.load_isd_hourly_temp_data(
        start, end, error_on_missing_years=False
    )
    assert ts.index[0] == start
    assert pd.isnull(ts[0])
    assert ts.index[-1] == end
    assert pd.notnull(ts[-1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import logging

from .__version__ import __title__, __description__, __url__, __version__
from .__version__ import __author__, __author_email__, __license__
from .__version__ import __copyright__
from .geo import get_lat_long_climate_zones, get_zcta_metadata, zcta_to_lat_long
from .database import build_metadata_db
from .exceptions import (
    EEWeatherError,
    UnrecognizedUSAFIDError,
    UnrecognizedZCTAError,
    ISDDataNotAvailableError,
    GSODDataNotAvailableError,
)
from .summaries import get_zcta_ids, get_isd_station_usaf_ids
from .ranking import rank_stations, combine_ranked_stations, select_station
from .stations import (
    ISDStation,
    get_isd_filenames,
    get_gsod_filenames,
    get_isd_station_metadata,
    get_isd_file_metadata,
    fetch_isd_raw_temp_data,
    fetch_isd_hourly_temp_data,
    fetch_isd_daily_temp_data,
    fetch_gsod_raw_temp_data,
    fetch_gsod_daily_temp_data,
    fetch_tmy3_hourly_temp_data,
    fetch_cz2010_hourly_temp_data,
    get_isd_hourly_temp_data_cache_key,
    get_isd_daily_temp_data_cache_key,
    get_gsod_daily_temp_data_cache_key,
    get_tmy3_hourly_temp_data_cache_key,
    get_cz2010_hourly_temp_data_cache_key,
    cached_isd_hourly_temp_data_is_expired,
    cached_isd_daily_temp_data_is_expired,
    cached_gsod_daily_temp_data_is_expired,
    validate_isd_hourly_temp_data_cache,
    validate_isd_daily_temp_data_cache,
    validate_gsod_daily_temp_data_cache,
    validate_tmy3_hourly_temp_data_cache,
    validate_cz2010_hourly_temp_data_cache,
    serialize_isd_hourly_temp_data,
    serialize_isd_daily_temp_data,
    serialize_gsod_daily_temp_data,
    serialize_tmy3_hourly_temp_data,
    serialize_cz2010_hourly_temp_data,
    deserialize_isd_hourly_temp_data,
    deserialize_isd_daily_temp_data,
    deserialize_gsod_daily_temp_data,
    deserialize_tmy3_hourly_temp_data,
    deserialize_cz2010_hourly_temp_data,
    read_isd_hourly_temp_data_from_cache,
    read_isd_daily_temp_data_from_cache,
    read_gsod_daily_temp_data_from_cache,
    read_tmy3_hourly_temp_data_from_cache,
    read_cz2010_hourly_temp_data_from_cache,
    write_isd_hourly_temp_data_to_cache,
    write_isd_daily_temp_data_to_cache,
    write_gsod_daily_temp_data_to_cache,
    write_tmy3_hourly_temp_data_to_cache,
    write_cz2010_hourly_temp_data_to_cache,
    destroy_cached_isd_hourly_temp_data,
    destroy_cached_isd_daily_temp_data,
    destroy_cached_gsod_daily_temp_data,
    destroy_cached_tmy3_hourly_temp_data,
    destroy_cached_cz2010_hourly_temp_data,
    load_isd_hourly_temp_data_cached_proxy,
    load_isd_daily_temp_data_cached_proxy,
    load_gsod_daily_temp_data_cached_proxy,
    load_tmy3_hourly_temp_data_cached_proxy,
    load_cz2010_hourly_temp_data_cached_proxy,
    load_isd_hourly_temp_data,
    load_isd_daily_temp_data,
    load_gsod_daily_temp_data,
    load_tmy3_hourly_temp_data,
    load_cz2010_hourly_temp_data,
    load_cached_isd_hourly_temp_data,
    load_cached_isd_daily_temp_data,
    load_cached_gsod_daily_temp_data,
    load_cached_tmy3_hourly_temp_data,
    load_cached_cz2010_hourly_temp_data,
)
from .visualization import plot_station_mapping, plot_station_mappings


def get_version():
    return __version__


# Set default logging handler to avoid "No handler found" warnings.
# will not work below py2.7
# workaround: https://docs.python.org/release/2.6/library/logging.html#configuring-logging-for-a-library
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np

from .connections import metadata_db_connection_proxy
from .exceptions import UnrecognizedUSAFIDError
from .stations import ISDStation

__all__ = ("plot_station_mapping", "plot_station_mappings")


def plot_station_mapping(
    target_latitude,
    target_longitude,
    isd_station,
    distance_meters,
    target_label="target",
):  # pragma: no cover
    """ Plots this mapping on a map."""
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Plotting requires matplotlib.")

    try:
        import cartopy.crs as ccrs
        import cartopy.feature as cfeature
        import cartopy.io.img_tiles as cimgt
    except ImportError:
        raise ImportError("Plotting requires cartopy.")

    lat, lng = isd_station.coords
    t_lat, t_lng = float(target_latitude), float(target_longitude)

    # fiture
    fig = plt.figure(figsize=(16, 8))

    # axes
    tiles = cimgt.Stamen("terrain-background")
    ax = plt.subplot(1, 1, 1, projection=tiles.crs)

    # offsets for labels
    x_max = max([lng, t_lng])
    x_min = min([lng, t_lng])
    x_diff = x_max - x_min

    y_max = max([lat, t_lat])
    y_min = min([lat, t_lat])
    y_diff = y_max - y_min

    xoffset = x_diff * 0.05
    yoffset = y_diff * 0.05

    # minimum
    left = x_min - x_diff * 0.5
    right = x_max + x_diff * 0.5
    bottom = y_min - y_diff * 0.3
    top = y_max + y_diff * 0.3

    width_ratio = 2.
    height_ratio = 1.

    if (right - left) / (top - bottom) > width_ratio / height_ratio:
        # too short
        goal = (right - left) * height_ratio / width_ratio
        diff = goal - (top - bottom)
        bottom = bottom - diff / 2.
        top = top + diff / 2.
    else:
        # too skinny
        goal = (top - bottom) * width_ratio / height_ratio
        diff = goal - (right - left)
        left = left - diff / 2.
        right = right + diff / 2.

    ax.set_extent([left, right, bottom, top])

    # determine zoom level
    # tile size at level 1 = 64 km
    # level 2 = 32 km, level 3 = 16 km, etc, i.e. 128/(2^n) km
    N_TILES = 600  # (how many tiles approximately fit in distance)
    km = distance_meters / 1000.0
    zoom_level = int(np.log2(128 * N_TILES / km))

    ax.add_image(tiles, zoom_level)

    # line between
    plt.plot(
        [lng, t_lng],
        [lat, t_lat],
        linestyle="-",
        dashes=[2, 2],
        transform=ccrs.Geodetic(),
    )

    # station
    ax.plot(lng, lat, "ko", markersize=7, transform=ccrs.Geodetic())

    # target
    ax.plot(t_lng, t_lat, "ro", markersize=7, transform=ccrs.Geodetic())

    # station label
    station_label = "{} ({})".format(isd_station.usaf_id, isd_station.name)
    ax.text(lng + xoffset, lat + yoffset, station_label, transform=ccrs.Geodetic())

    # target label
    ax.text(t_lng + xoffset, t_lat + yoffset, target_label, transform=ccrs.Geodetic())

    # distance labels
    mid_lng = (lng + t_lng) / 2
    mid_lat = (lat + t_lat) / 2
    dist_text = "{:.01f} km".format(km)
    ax.text(mid_lng + xoffset, mid_lat + yoffset, dist_text, transform=ccrs.Geodetic())

    plt.show()


def plot_station_mappings(mapping_results):  # pragma: no cover
    """ Plot a list of mapping results on a map.

    Requires matplotlib and cartopy.

    Parameters
    ----------
    mapping_results : list of MappingResult objects
        Mapping results to plot
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Plotting requires matplotlib.")

    try:
        import cartopy.crs as ccrs
        import cartopy.feature as cfeature
    except ImportError:
        raise ImportError("Plotting requires cartopy.")

    lats = []
    lngs = []
    t_lats = []
    t_lngs = []
    n_discards = 0
    for mapping_result in mapping_results:
        if not mapping_result.is_empty():
            lat, lng = mapping_result.isd_station.coords
            t_lat, t_lng = map(float, mapping_result.target_coords)
            lats.append(lat)
            lngs.append(lng)
            t_lats.append(t_lat)
            t_lngs.append(t_lng)
        else:
            n_discards += 1

    print("Discarded {} empty mappings".format(n_discards))

    # figure
    fig = plt.figure(figsize=(60, 60))

    # axes
    ax = plt.subplot(1, 1, 1, projection=ccrs.Mercator())

    # offsets for labels
    all_lngs = lngs + t_lngs
    all_lats = lats + t_lats
    x_max = max(all_lngs)  # lists
    x_min = min(all_lngs)
    x_diff = x_max - x_min

    y_max = max(all_lats)
    y_min = min(all_lats)
    y_diff = y_max - y_min

    # minimum
    x_pad = 0.1 * x_diff
    y_pad = 0.1 * y_diff
    left = x_min - x_pad
    right = x_max + x_pad
    bottom = y_min - y_pad
    top = y_max + y_pad

    width_ratio = 2.
    height_ratio = 1.

    if (right - left) / (top - bottom) > height_ratio / width_ratio:
        # too short
        goal = (right - left) * height_ratio / width_ratio
        diff = goal - (top - bottom)
        bottom = bottom - diff / 2.
        top = top + diff / 2.
    else:
        # too skinny
        goal = (top - bottom) * width_ratio / height_ratio
        diff = goal - (right - left)
        left = left - diff / 2.
        right = right + diff / 2.

    left = max(left, -179.9)
    right = min(right, 179.9)
    bottom = max([bottom, -89.9])
    top = min([top, 89.9])

    ax.set_extent([left, right, bottom, top])

    # OCEAN
    ax.add_feature(
        cfeature.NaturalEarthFeature(
            "physical",
            "ocean",
            "50m",
            edgecolor="face",
            facecolor=cfeature.COLORS["water"],
        )
    )

    # LAND
    ax.add_feature(
        cfeature.NaturalEarthFeature(
            "physical",
            "land",
            "50m",
            edgecolor="face",
            facecolor=cfeature.COLORS["land"],
        )
    )

    # BORDERS
    ax.add_feature(
        cfeature.NaturalEarthFeature(
            "cultural",
            "admin_0_boundary_lines_land",
            "50m",
            edgecolor="black",
            facecolor="none",
        )
    )

    # LAKES
    ax.add_feature(
        cfeature.NaturalEarthFeature(
            "physical",
            "lakes",
            "50m",
            edgecolor="face",
            facecolor=cfeature.COLORS["water"],
        )
    )

    # COASTLINE
    ax.add_feature(
        cfeature.NaturalEarthFeature(
            "physical", "coastline", "50m", edgecolor="black", facecolor="none"
        )
    )

    # lines between
    # for lat, t_lat, lng, t_lng in zip(lats, t_lats, lngs, t_lngs):
    ax.plot(
        [lngs, t_lngs],
        [lats, t_lats],
        color="k",
        linestyle="-",
        transform=ccrs.Geodetic(),
        linewidth=0.3,
    )

    # stations
    ax.plot(lngs, lats, "bo", markersize=1, transform=ccrs.Geodetic())

    plt.title("Location to weather station mapping")

    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from io import BytesIO
import pkg_resources
import re
import tempfile

from eeweather.cache import KeyValueStore


def write_isd_file(bytes_string):
    with pkg_resources.resource_stream("eeweather.resources", "ISD.gz") as f:
        bytes_string.write(f.read())


def write_tmy3_file():
    data = pkg_resources.resource_string("eeweather.resources", "722880TYA.CSV")
    return data.decode("ascii")


def write_cz2010_file():
    data = pkg_resources.resource_string("eeweather.resources", "722880_CZ2010.CSV")
    return data.decode("ascii")


def write_missing_isd_file(bytes_string):
    with pkg_resources.resource_stream("eeweather.resources", "ISD-MISSING.gz") as f:
        bytes_string.write(f.read())


def write_nan_isd_file(bytes_string):
    with pkg_resources.resource_stream("eeweather.resources", "ISD-NAN.gz") as f:
        bytes_string.write(f.read())


def write_gsod_file(bytes_string):
    with pkg_resources.resource_stream("eeweather.resources", "GSOD.op.gz") as f:
        bytes_string.write(f.read())


def write_missing_gsod_file(bytes_string):
    with pkg_resources.resource_stream(
        "eeweather.resources", "GSOD-MISSING.op.gz"
    ) as f:
        bytes_string.write(f.read())


def mock_request_text_tmy3(url):
    match_url = (
        "http://rredc.nrel.gov/solar/old_data/nsrdb/"
        "1991-2005/data/tmy3/722880TYA.CSV"
    )
    if re.match(match_url, url):
        return write_tmy3_file()


def mock_request_text_cz2010(url):
    match_url = "https://storage.googleapis.com/oee-cz2010/csv/722880_CZ2010.CSV"

    if re.match(match_url, url):
        return write_cz2010_file()


class MockNOAAFTPConnectionProxy:
    def read_file_as_bytes(self, filename):
        bytes_string = BytesIO()

        if re.match("/pub/data/noaa/2007/722874-93134-2007.gz", filename):
            write_isd_file(bytes_string)
        elif re.match("/pub/data/noaa/2006/722874-93134-2006.gz", filename):
            write_missing_isd_file(bytes_string)
        elif re.match("/pub/data/noaa/2013/994035-99999-2013.gz", filename):
            write_nan_isd_file(bytes_string)
        elif re.match("/pub/data/gsod/2007/722874-93134-2007.op.gz", filename):
            write_gsod_file(bytes_string)
        elif re.match("/pub/data/gsod/2006/722874-93134-2006.op.gz", filename):
            write_missing_gsod_file(bytes_string)

        bytes_string.seek(0)
        return bytes_string


class MockKeyValueStoreProxy:
    def __init__(self):
        # create a new test store in a temporary folder
        self.store = KeyValueStore("sqlite:///{}/cache.db".format(tempfile.mkdtemp()))

    def get_store(self):
        return self.store



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2018 Open Energy Efficiency, Inc.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from eeweather.validation import valid_zcta_or_raise, valid_usaf_id_or_raise
from eeweather.exceptions import UnrecognizedZCTAError, UnrecognizedUSAFIDError
import pytest


def test_valid_zcta_or_raise_valid():
    assert valid_zcta_or_raise("90210") is True


def test_valid_zcta_or_raise_raise_error():
    with pytest.raises(UnrecognizedZCTAError) as excinfo:
        valid_zcta_or_raise("INVALID")
    assert excinfo.value.value == "INVALID"


def test_valid_usaf_id_or_raise_valid():
    assert valid_usaf_id_or_raise("722880") is True


def test_valid_usaf_id_or_raise_raise_error():
    with pytest.raises(UnrecognizedUSAFIDError) as excinfo:
        valid_usaf_id_or_raise("INVALID")
    assert excinfo.value.value == "INVALID"
